from __future__ import annotations

import os

import pytest

from src.scheduling_and_allocation.export.registrars_office import registrars_output
from src.scheduling_and_allocation.model import TimeSlot, Course, Section, Block, Lab, Teacher, ConflictType, Schedule


@pytest.fixture
def schedule1():
    schedule = Schedule()
    schedule.add_update_course("abc","ABC of cooking")

def test_reates_output_file():
    # prepare
    schedule = Schedule()

    # execute
    filename="dummy.csv"
    if os.path.exists(filename):
        os.remove(filename)
    registrars_output(schedule, filename)

    # validate
    fh = open(filename,"r")
    rows = fh.readlines()
    assert len(rows) == 1, "header line was created"

def test_ignores_non_allocated_course():
    # prepare
    schedule = Schedule()
    c_abc = schedule.add_update_course("abc", "ABC of cooking")
    s_abc_1 = c_abc.add_section()
    s_abc_1.add_teacher(Teacher("Fay","Runaway"))
    c_000 = schedule.add_update_course("000", "Dept Meeting",needs_allocation=False)
    s_000_1 = c_000.add_section()
    s_000_1.add_teacher(Teacher("Fay","Runaway"))
    s_000_1.add_teacher(Teacher("Babe","Ruth"))


    # execute
    filename="dummy.csv"
    if os.path.exists(filename):
        os.remove(filename)
    registrars_output(schedule, filename)

    # validate
    fh = open(filename,"r")
    rows = fh.readlines()
    assert len(rows) == 2, "only once course is added to csv file"
    data = rows[1].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"


def test_no_teacher_within_block_and_fields_are_good():
    # prepare
    schedule = Schedule()
    c_abc = schedule.add_update_course("abc", "ABC of cooking")
    s_abc_1 = c_abc.add_section()
    b_s_abc_1 = s_abc_1.add_block(1,8.5,1.5)

    # execute
    filename="dummy.csv"
    if os.path.exists(filename):
        os.remove(filename)
    registrars_output(schedule, filename)

    # validate
    fh = open(filename,"r")
    rows = fh.readlines()
    assert len(rows) == 2, "only once course is added to csv file"
    data = rows[1].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == "830"
    assert data[4] == "1000"
    assert data[5] == "Monday"
    assert data[6] == ""
    assert data[7] == ""

def test_no_teacher_no_block_and_fields_are_good():
    # prepare
    schedule = Schedule()
    c_abc = schedule.add_update_course("abc", "ABC of cooking")
    s_abc_1 = c_abc.add_section()

    # execute
    filename="dummy.csv"
    if os.path.exists(filename):
        os.remove(filename)
    registrars_output(schedule, filename)

    # validate
    fh = open(filename,"r")
    rows = fh.readlines()
    assert len(rows) == 2, "only once course is added to csv file"
    data = rows[1].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == ""
    assert data[4] == ""
    assert data[5] == ""
    assert data[6] == ""
    assert data[7] == ""

def test_single_teacher_within_block_and_fields_are_good():
    # prepare
    schedule = Schedule()
    c_abc = schedule.add_update_course("abc", "ABC of cooking")
    s_abc_1 = c_abc.add_section()
    b_s_abc_1 = s_abc_1.add_block(2,8.5,1.5)
    b_s_abc_2 = s_abc_1.add_block(3,11.5,1.5)
    b_s_abc_1.add_teacher(Teacher("Noah","Way"))
    b_s_abc_2.add_teacher(Teacher("Hugh","Bet"))

    # execute
    filename="dummy.csv"
    if os.path.exists(filename):
        os.remove(filename)
    registrars_output(schedule, filename)

    # validate
    fh = open(filename,"r")
    rows = fh.readlines()
    assert len(rows) == 3, "only once course, but two blocks is added to csv file"
    data = rows[1].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == "830"
    assert data[4] == "1000"
    assert data[5] == "Tuesday"
    assert data[6] == "Way"
    assert data[7] == "Noah"

    data = rows[2].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == "1130"
    assert data[4] == "1300"
    assert data[5] == "Wednesday"
    assert data[6] == "Bet"
    assert data[7] == "Hugh"

def test_single_teacher_no_block_and_fields_are_good():
    # prepare
    schedule = Schedule()
    c_abc = schedule.add_update_course("abc", "ABC of cooking")
    s_abc_1 = c_abc.add_section()
    s_abc_1.set_teacher_allocation(Teacher("Teddy","Bear"),3)
    # b_s_abc_1 = s_abc_1.add_block(2,8.5,1.5)
    # b_s_abc_2 = s_abc_1.add_block(3,11.5,1.5)
    # b_s_abc_1.add_teacher(Teacher("Noah","Way"))
    # b_s_abc_2.add_teacher(Teacher("Hugh","Bet"))

    # execute
    filename="dummy.csv"
    if os.path.exists(filename):
        os.remove(filename)
    registrars_output(schedule, filename)

    # validate
    fh = open(filename,"r")
    rows = fh.readlines()
    assert len(rows) == 2, "only once course, no blocks is added to csv file"
    data = rows[1].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == ""
    assert data[4] == ""
    assert data[5] == ""
    assert data[6] == "Bear"
    assert data[7] == "Teddy"

def test_multiple_teachers_within_block_and_fields_are_good():
    # prepare
    schedule = Schedule()
    c_abc = schedule.add_update_course("abc", "ABC of cooking")
    s_abc_1 = c_abc.add_section()
    # s_abc_1.set_teacher_allocation(Teacher("Teddy","Bear"),3)
    b_s_abc_1 = s_abc_1.add_block(2,8.5,1.5)
    b_s_abc_1.add_teacher(Teacher("Noah","Way"))
    b_s_abc_1.add_teacher(Teacher("Hugh","Bet"))

    # execute
    filename="dummy.csv"
    if os.path.exists(filename):
        os.remove(filename)
    registrars_output(schedule, filename)

    # validate
    fh = open(filename,"r")
    rows = fh.readlines()
    assert len(rows) == 3, "only once course, two blocks is added to csv file"
    data = rows[1].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == "830"
    assert data[4] == "1000"
    assert data[5] == "Tuesday"
    assert data[6] == "Bet"
    assert data[7] == "Hugh"

    data = rows[2].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == "830"
    assert data[4] == "1000"
    assert data[5] == "Tuesday"
    assert data[6] == "Way"
    assert data[7] == "Noah"

def test_multiple_teachers_no_block_and_fields_are_good():
    # prepare
    schedule = Schedule()
    c_abc = schedule.add_update_course("abc", "ABC of cooking")
    s_abc_1 = c_abc.add_section()
    s_abc_1.set_teacher_allocation(Teacher("Teddy","Bear"),1.5)
    s_abc_1.set_teacher_allocation(Teacher("Toodle","Loo"),1.5)

    # b_s_abc_1 = s_abc_1.add_block(2,8.5,1.5)
    # b_s_abc_1.add_teacher(Teacher("Noah","Way"))
    # b_s_abc_1.add_teacher(Teacher("Hugh","Bet"))

    # execute
    filename="dummy.csv"
    if os.path.exists(filename):
        os.remove(filename)
    registrars_output(schedule, filename)

    # validate
    fh = open(filename,"r")
    rows = fh.readlines()
    assert len(rows) == 3, "only once course, zero blocks, two teachers, is added to csv file"
    data = rows[1].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == ""
    assert data[4] == ""
    assert data[5] == ""
    assert data[6] == "Bear"
    assert data[7] == "Teddy"

    data = rows[2].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == ""
    assert data[4] == ""
    assert data[5] == ""
    assert data[6] == "Loo"
    assert data[7] == "Toodle"

def test_zero_lab_within_block_and_fields_are_good():
    # prepare
    schedule = Schedule()
    c_abc = schedule.add_update_course("abc", "ABC of cooking")
    s_abc_1 = c_abc.add_section()
    b_s_abc_1 = s_abc_1.add_block(2,8.5,1.5)

    # execute
    filename="dummy.csv"
    if os.path.exists(filename):
        os.remove(filename)
    registrars_output(schedule, filename)

    # validate
    fh = open(filename,"r")
    rows = list(map(str.rstrip,fh.readlines()))
    assert len(rows) == 2, "only once course, one blocks is added to csv file"
    data = rows[1].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == "830"
    assert data[4] == "1000"
    assert data[5] == "Tuesday"
    assert data[6] == ""
    assert data[7] == ""
    assert data[8] == ""
    assert data[9] == ""


# handles zero lab with no block
def test_zero_lab_no_block_and_fields_are_good():
    # prepare
    schedule = Schedule()
    c_abc = schedule.add_update_course("abc", "ABC of cooking")
    s_abc_1 = c_abc.add_section()
    #b_s_abc_1 = s_abc_1.add_block(2,8.5,1.5)

    # execute
    filename="dummy.csv"
    if os.path.exists(filename):
        os.remove(filename)
    registrars_output(schedule, filename)

    # validate
    fh = open(filename,"r")
    rows = list(map(str.rstrip,fh.readlines()))
    assert len(rows) == 2, "only once course, one blocks is added to csv file"
    data = rows[1].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == ""
    assert data[4] == ""
    assert data[5] == ""
    assert data[6] == ""
    assert data[7] == ""
    assert data[8] == ""
    assert data[9] == ""


def test_single_lab_within_block_and_fields_are_good():
    # prepare
    schedule = Schedule()
    c_abc = schedule.add_update_course("abc", "ABC of cooking")
    s_abc_1 = c_abc.add_section()
    b_s_abc_1 = s_abc_1.add_block(2,8.5,1.5)
    b_s_abc_1.add_lab(Lab("P322"))

    # execute
    filename="dummy.csv"
    if os.path.exists(filename):
        os.remove(filename)
    registrars_output(schedule, filename)

    # validate
    fh = open(filename,"r")
    rows = list(map(str.rstrip,fh.readlines()))
    assert len(rows) == 2, "only once course, one blocks is added to csv file"
    data = rows[1].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == "830"
    assert data[4] == "1000"
    assert data[5] == "Tuesday"
    assert data[6] == ""
    assert data[7] == ""
    assert data[8] == "P322"
    assert data[9] == ""

# handles multiple labs with block
def test_multiple_labs_with_blocks():
    # prepare
    schedule = Schedule()
    c_abc = schedule.add_update_course("abc", "ABC of cooking")
    s_abc_1 = c_abc.add_section()
    b_s_abc_1 = s_abc_1.add_block(2,8.5,1.5)
    b_s_abc_1.add_lab(Lab("P322"))
    b_s_abc_1.add_lab(Lab("P107"))
    b_s_abc_1.add_lab(Lab("P325"))

    # execute
    filename="dummy.csv"
    if os.path.exists(filename):
        os.remove(filename)
    registrars_output(schedule, filename)

    # validate
    fh = open(filename,"r")
    rows = list(map(str.rstrip,fh.readlines()))
    assert len(rows) == 2, "only once course, one blocks is added to csv file"
    data = rows[1].split(",")
    assert data[1] == "abc", "the course added to the csv file is abd"
    assert data[2] == "1", "section number"
    assert data[3] == "830"
    assert data[4] == "1000"
    assert data[5] == "Tuesday"
    assert data[6] == ""
    assert data[7] == ""
    assert data[8] == "P107"
    assert data[9] == "P322 P325"




"""
AddRemoveTk

Creates two list boxes where you can transfer items from one list to another, just by clicking on an item

"""
import tkinter as tk

from ..modified_tk import Scrolled
from functools import partial

# ====================================================================================================================
# Move objects between two lists
# ====================================================================================================================
class AddRemoveTk:
    def __init__(self, frame: tk.Frame,
                 add_list: list,
                 remove_list: list,
                 add_text="Add to",
                 remove_text="Remove from",
                 height = 10):
        """
        :param frame: the frame to put the widgets in
        :param add_list: the list of objects that can be added
        :param remove_list: the list of objects that can be removed
        :param add_text: text over the 'add' list
        :param remove_text: text over the 'remove' list
        """
        self.frame = frame
        self.add_text = add_text
        self.remove_text = remove_text
        self.add_listbox = None
        self.remove_listbox = None
        self.adds = add_list
        self.removes = remove_list

        f = tk.Frame(self.frame)
        f.grid(column=0, stick='nsew', row=0)
        tk.Label(f, text=self.add_text).pack()
        sf = tk.Frame(f)
        sf.pack(fill='both', expand=1)
        s: Scrolled = Scrolled(sf, 'Listbox', scrollbars='oe', height=height)
        self.add_listbox = s.widget
        self.add_listbox.configure(borderwidth="5", relief="ridge")
        s.widget.bind('<Button-1>', partial(self._cmd_click, 'add'))

        f = tk.Frame(self.frame)
        f.grid(column=1, stick='nsew', row=0)
        tk.Label(f, text=self.remove_text).pack()
        sf = tk.Frame(f)
        sf.pack(fill='both', expand=1)
        s: Scrolled = Scrolled(sf, 'Listbox', scrollbars='oe', height=height)
        self.remove_listbox = s.widget
        self.remove_listbox.configure(borderwidth="5", relief="ridge")
        s.widget.bind('<Button-1>', partial(self._cmd_click, 'remove'))

        self.refresh()

    # ----------------------------------------------------------------------------------------------------------------
    # move object from 'remove' list to 'add' list
    # ----------------------------------------------------------------------------------------------------------------
    def to_add_function(self, obj):
        self.removes.append(obj)
        try:
            self.removes.sort()
        except ValueError:
            pass
        self.adds.remove(obj)

    # ----------------------------------------------------------------------------------------------------------------
    # move object from 'add' list to 'remove' list
    # ----------------------------------------------------------------------------------------------------------------
    def to_remove_function(self, obj):
        self.adds.append(obj)
        try:
            self.adds.sort()
        except ValueError:
            pass
        self.removes.remove(obj)

    # ----------------------------------------------------------------------------------------------------------------
    # what to do when a list box is clicked (... move object)
    # ----------------------------------------------------------------------------------------------------------------
    def _cmd_click(self, which: str, e: tk.Event):
        """
        move object from one list to another
        :param which: which is the 'starting' list
        :param e: the tk event (which can give us the nearest widget to the mouse click)
        :return:
        """
        if which == "add":
            widget = self.add_listbox
            index = widget.nearest(e.y)
            if 0 <= index < len(self.adds):
                self.to_add_function(self.adds[index])
        else:
            widget = self.remove_listbox
            index = widget.nearest(e.y)
            if 0 <= index < len(self.removes):
                self.to_remove_function(self.removes[index])
        self.refresh()

    # ----------------------------------------------------------------------------------------------------------------
    # refrest the two list boxes
    # ----------------------------------------------------------------------------------------------------------------
    def refresh(self):
        """updates the two list boxes"""
        self.add_listbox.delete(0, "end")
        self.remove_listbox.delete(0, "end")
        for item in self.adds:
            self.add_listbox.insert("end", str(item))
        for item in self.removes:
            self.remove_listbox.insert("end", str(item))



import tkinter as tk
from tkinter import messagebox
from functools import partial
"""
Functions to help with number validations, especially with entry widgets
"""

# ================================================================================================================
# register the functions for 'entry' widget validations
# ================================================================================================================
def register_number_funcs(frame:tk.Frame):
    """register the functions for 'entry' widget validations
    :param frame: any Tk object
    """
    tk_is_int = frame.winfo_toplevel().register(partial(_is_int, frame))
    tk_is_float = frame.winfo_toplevel().register(partial(_is_float, frame))
    return tk_is_int, tk_is_float

# ================================================================================================================
# enter a float
# ================================================================================================================
def entry_float(frame: tk.Frame, textvariable: tk.StringVar, **kwargs ) -> tk.Entry:
    """
    An entry widget that only accepts floats (note... '' and '.' would be valid, so you still need to check later)
    :param frame:
    :param textvariable:
    :return: the entry widget with validation
    """
    _,tk_is_float = register_number_funcs(frame)

    if "justify" not in kwargs:
        kwargs["justify"] = "right"

    return tk.Entry(frame,
                           textvariable=textvariable,
                           validate='key',
                           validatecommand=(tk_is_float, '%P', '%s'),
                            **kwargs)

# ================================================================================================================
# enter an integer
# ================================================================================================================
def entry_int(frame: tk.Frame, textvariable: tk.StringVar ) -> tk.Entry:
    """
    An entry widget that only accepts ints (note... '' and would be valid, so you still need to check later)
    :param frame:
    :param textvariable:
    :return: the entry widget with validation
    """
    tk_is_int,_ = register_number_funcs(frame)
    return tk.Entry(frame,
                           textvariable=textvariable,
                           validate='key',
                           validatecommand=(tk_is_int, '%P', '%s'),
                           justify='right')

# ================================================================================================================
# is it an int... or the start of a valid int (i.e. "" returns true)
# - used to validate data as it is being entered into an 'entry' widget
# ================================================================================================================
def _is_int(frame: tk.Frame, number: str, *_) -> bool:
    """
    Validation for the string that is currently in an Entry widget
    :param number: the number_of_students that would result if this validation returns True
    :return: is this a valid int, or the start of a valid int
    """
    if number == "":
        return True
    try:
        int(number)
        return True
    except ValueError:
        frame.winfo_toplevel().bell()
        return False


# ================================================================================================================
# is it an  float... or the start of a valid float (i.e. "", "." returns true)
# - used to validate data as it is being entered into an 'entry' widget
# ================================================================================================================
def _is_float(frame: tk.Frame, number: str, _: str) -> bool:
    """
    Validation for the string that is currently in an Entry widget
    :param number: the number that would result if this validation returns True
    :return: is this a valid float, or the start of a valid float
    """
    number = number.strip()
    if number == "" or number == ".":
        return True
    try:
        float(number)
        return True
    except ValueError:
        frame.winfo_toplevel().bell()
        return False

# ================================================================================================================
# is this really a float?
# ================================================================================================================
def validate_float(number:str, title, msg)-> bool:
    """
    is this number a float
    :param number: the str version of the number to test
    :param title: dialog box title
    :param msg: dialog box message
    :return: True/False
    """
    try:
        float(number)
    except ValueError:
        messagebox.showerror(title, msg)
        return False
    return True

# ================================================================================================================
# is this really a int?
# ================================================================================================================
def validate_int(number:str, title, msg)-> bool:
    """
    is this number a int
    :param number: the str version of the number to test
    :param title: dialog box title
    :param msg: dialog box message
    :return: True/False
    """
    try:
        int(number)
    except ValueError:
        messagebox.showerror(title, msg)
        return False
    return True




from src.scheduling_and_allocation.model import TimeSlot, WeekDay, \
    MINIMUM_DURATION, DEFAULT_DAY, DEFAULT_START, DEFAULT_DURATION, MINUTE_BLOCK_SIZE, MAXIMUM_DURATION, MAX_END_TIME


def test_defaults():
    slot = TimeSlot()
    assert slot.day == DEFAULT_DAY
    assert slot.start == DEFAULT_START
    assert slot.duration == DEFAULT_DURATION
    assert slot.movable


def test_calculate_end_time():
    slot = TimeSlot(WeekDay.Tuesday, start=13.25, duration=1.5)
    assert slot.day == WeekDay.Tuesday
    assert slot.start == 13.25
    assert slot.duration == 1.5
    assert slot.end == 14.75


def test_equality():
    slot1 = TimeSlot(WeekDay.Tuesday, start=13.25, duration=1.5)
    slot2 = TimeSlot(WeekDay.Tuesday, start=13.25, duration=1.5)
    assert slot1 == slot2


def test_hashable():
    slot1 = TimeSlot(WeekDay.Tuesday, start=13.25, duration=1.5)
    slot2 = TimeSlot(WeekDay.Tuesday, start=13.25, duration=1.5)
    s = {slot1, slot2}
    assert len(s) == 1


def test_sortable():
    slot1 = TimeSlot(WeekDay.Tuesday, start=13.25, duration=1.5)
    slot2 = TimeSlot(WeekDay.Monday, start=13.25, duration=1.5)
    s = [slot1, slot2]
    s.sort()
    assert s[0] == slot2
    assert s[1] == slot1


def test_duration_not_allowed_less_than_zero():
    slot = TimeSlot(WeekDay.Tuesday, start=13.25, duration=-3)
    assert abs(slot.duration - MINUTE_BLOCK_SIZE / 60) < 0.001


def test_duration_not_allowed_greater_than_eight():
    slot = TimeSlot(WeekDay.Tuesday, start=13.5, duration=MAXIMUM_DURATION + .5)
    assert slot.duration == MAXIMUM_DURATION


def test_duration_rounded_up_to_minimum_half_hour():
    slot = TimeSlot(WeekDay.Tuesday, 13.5, duration=0.1)
    assert slot.duration == MINIMUM_DURATION


def test_duration_setter_changes_end():
    slot = TimeSlot(WeekDay.Tuesday, 13.5, duration=1)
    assert slot.end == 14.5
    slot.duration = 3
    assert slot.end == 16.5


def test_no_snapping_required():
    slot = TimeSlot(start=13)
    assert not slot.snap_to_time()


def test_snap_to_time():
    slot = TimeSlot(start=13.25)
    slot.snap_to_time()
    expected_start =13
    assert slot.start == expected_start


def test_snap_to_time_adjust_duration():
    slot = TimeSlot(duration=1.2, start=13)
    slot.snap_to_time()
    assert slot.duration == 1


def test_snap_to_time_bad_value_to_minimum():
    slot = TimeSlot(start=5)
    slot.snap_to_time()
    expected_start = 8
    print()
    print(slot)
    assert slot.start == expected_start


def test_snap_to_time_bad_value_to_maximum():
    slot = TimeSlot(start=19)
    slot.snap_to_time()
    assert slot.start + slot.duration <= MAX_END_TIME


def test_snap_to_day_with_args():
    """Verifies that TimeSlot's snap_to_day method adjusts its day property to the correct value
    when it is outside a specified date range. """
    slot = TimeSlot()
    slot.day = WeekDay.Friday
    assert slot.snap_to_day(slot.day.value, WeekDay.Tuesday, WeekDay.Wednesday)
    assert slot.day == WeekDay.Wednesday


def test_snap_to_day_with_fractional_day():
    slot = TimeSlot()
    slot.day = WeekDay.Friday
    assert slot.snap_to_day(2.3)
    assert slot.day == WeekDay.Tuesday


def test_snap_to_day_returns_false_no_change():
    slot = TimeSlot()
    slot.day = WeekDay.Tuesday
    assert not slot.snap_to_day(2)
    assert slot.day == WeekDay.Tuesday


def test_conflicts_time1():
    """Verifies that conflicts_time works as intended."""
    slot1 = TimeSlot(WeekDay.Monday,8, 1.5)
    slot2 = TimeSlot(WeekDay.Monday, 9, 1)
    assert slot1.conflicts_time(slot2)


def test_conflicts_time2():
    """Verifies that conflicts_time works as intended."""
    slot1 = TimeSlot(WeekDay.Monday, 8, 1)
    slot2 = TimeSlot(WeekDay.Monday, 9, 1)
    assert not slot1.conflicts_time(slot2)


def test_conflicts_time3():
    """Verifies that conflicts_time works as intended."""
    slot1 = TimeSlot(WeekDay.Monday,9.5, 1)
    slot2 = TimeSlot(WeekDay.Monday,9, 1)
    assert slot1.conflicts_time(slot2)


def test_conflicts_time4():
    """Verifies that conflicts_time works as intended."""
    slot1 = TimeSlot(WeekDay.Monday, 9.5, 1)
    slot2 = TimeSlot(WeekDay.Monday, 10.5, 1)
    assert not slot1.conflicts_time(slot2)


def test_conflicts_time_different_days():
    """Verifies that conflicts_time registers no conflict when two TimeSlots are on different
    days. """
    slot1 = TimeSlot(WeekDay.Tuesday, 9.5, 1)
    slot2 = TimeSlot(WeekDay.Monday, 9, 1)
    assert slot1.conflicts_time(slot2) is False

def test_conflicts_time_one_inside_another1():
    slot1 = TimeSlot(WeekDay.Tuesday, 9.5, 1)
    slot2 = TimeSlot(WeekDay.Tuesday, 9.5, 4)
    assert slot1.conflicts_time(slot2) is True
    assert slot2.conflicts_time(slot1) is True

def test_conflicts_time_one_inside_another2():
    slot1 = TimeSlot(WeekDay.Tuesday, 9.5, 4)
    slot2 = TimeSlot(WeekDay.Tuesday, 10, 1)
    assert slot1.conflicts_time(slot2) is True
    assert slot2.conflicts_time(slot1) is True

def test_conflicts_time_one_inside_another3():
    slot1 = TimeSlot(WeekDay.Tuesday, 9.5, 4)
    slot2 = TimeSlot(WeekDay.Tuesday, 9.5, 4)
    assert slot1.conflicts_time(slot2) is True
    assert slot2.conflicts_time(slot1) is True



from __future__ import annotations

from filecmp import cmp
from os import path

import pytest

from src.scheduling_and_allocation.model import Schedule


def test_read():
    """stupidest simplest test whatsoever"""

    # reading doesn't crash
    # TODO: not a good test
    schedule = Schedule(path.dirname(__file__) + "/data_test_good_input.csv")



def test_read_write():
    """stupidest simplest test whatsoever"""

    # logic is that what it reads in is saved in the schedule object, so it should print out the
    # same info to the output file

    # TODO: not a good test
    schedule = Schedule(path.dirname(__file__) + "/data_test_good_input.csv")
    teachers_text = list()
    for t in schedule.teachers():
        teachers_text.append(schedule.teacher_details(t))

    schedule.write_file(path.dirname(__file__) + "/data_test_output.csv")
    assert (cmp(path.dirname(__file__) + "/data_test_good_input.csv", path.dirname(__file__) + "/data_test_output.csv"))


# test_read_write()


#schedule = s.Schedule(path.dirname(__file__) + "/test.csv")
def test_schedule_actually_has_stuff_in_it():
    """smoke test to see if schedule is actually populated with data - not validating data though"""
    schedule = Schedule(path.dirname(__file__) + "/data_test_good_input.csv")
    assert len(schedule._teachers) != 0
    assert len(schedule.courses()) != 0
    assert len(schedule.labs()) != 0
    assert len(schedule._streams) != 0

def test_blocks_movable_is_correct():
    schedule = Schedule(path.dirname(__file__) + "/data_test_good_input.csv")
    teacher = schedule.get_teacher_by_number("Bultena_Sandy")
    blocks = schedule.get_blocks_for_teacher(teacher)
    for b in blocks:
        assert b.movable()




"""
# ============================================================================
# Gathers info for main menus and toolbars for scheduler, and gives them default commands
# - to set the commands, use methods 'set_menu_event_handler'
#
# NOTE: this does not actually create the gui menu, it just gathers the info
# ============================================================================
"""
from __future__ import annotations

from functools import partial
from typing import Callable, Literal, get_args
from ..gui_generics.menu_and_toolbars import MenuItem, MenuType, ToolbarItem
from ..model import SemesterType

MAIN_MENU_EVENT_HANDLER_NAMES_ALLOCATION = Literal[
    "file_new",
    "file_open",
    "file_save",
    "file_save_as",
    "file_exit",
    "auto_save",
]


MAIN_MENU_EVENT_HANDLERS_ALLOCATION: dict[MAIN_MENU_EVENT_HANDLER_NAMES_ALLOCATION, Callable[[SemesterType], None]] = {}
for event_name in get_args(MAIN_MENU_EVENT_HANDLER_NAMES_ALLOCATION):
    MAIN_MENU_EVENT_HANDLERS_ALLOCATION[event_name] = lambda semester: print(f"'{event_name}' selected. ")


def set_menu_event_handler_allocation(name: MAIN_MENU_EVENT_HANDLER_NAMES_ALLOCATION,
                                      handler: Callable[[SemesterType], None]):
    MAIN_MENU_EVENT_HANDLERS_ALLOCATION[name] = handler


def main_menu_allocation(semesters:list[SemesterType]) -> tuple[list[str], dict[str, ToolbarItem], list[MenuItem]]:

    menu = list()
    # -----------------------------------------------------------------------------------------
    # File menu
    # -----------------------------------------------------------------------------------------
    file_menu = MenuItem(name='file', menu_type=MenuType.Cascade, label='File')
    for semester in semesters:
        file_menu.add_child(MenuItem(name=f"new_{semester.name}", menu_type=MenuType.Command,
                                 label=f'New {semester.name} Schedule',
                                 command=partial( MAIN_MENU_EVENT_HANDLERS_ALLOCATION["file_new"],semester)
                                 )
                        )
        file_menu.add_child(MenuItem(name=f'open_{semester.name}', menu_type=MenuType.Command,
                                     label=f'Open {semester.name} Schedule',
                                     command=partial( MAIN_MENU_EVENT_HANDLERS_ALLOCATION["file_open"],semester)
                                     )
                            )

        file_menu.add_child(MenuItem(menu_type=MenuType.Separator))

    file_menu.add_child(MenuItem(name=f'save', menu_type=MenuType.Command,
                                 label=f'Save Schedule',
                                 command=partial( MAIN_MENU_EVENT_HANDLERS_ALLOCATION["file_save"],None)
                                 )
                        )
    file_menu.add_child(MenuItem(menu_type=MenuType.Separator))

    file_menu.add_child(MenuItem(menu_type=MenuType.Command,
                                 label='Exit',
                                 accelerator='Ctrl-e',
                                 command=partial( MAIN_MENU_EVENT_HANDLERS_ALLOCATION["file_exit"],None)
                                 )
                        )
    # -----------------------------------------------------------------------------------------
    # Auto Save - coding is in the main-tk
    # -----------------------------------------------------------------------------------------



    # -----------------------------------------------------------------------------------------
    # _toolbar
    # -----------------------------------------------------------------------------------------
    toolbar_info = dict()
    toolbar_order = []

    # return list of top level menu items
    menu.append(file_menu)

    return toolbar_order, toolbar_info, menu




import os.path
from functools import partial
from PIL import Image, ImageTk

from typing import Callable, Optional, Literal, get_args

from ..Utilities.Preferences import Preferences

from .main_pages_tk import MainPageBaseTk
import tkinter as tk
from ..model.enums import SemesterType

BUTTON_WIDTH = 50
MAX_LEN_OF_DISPLAYED_FILENAME = 60

# ============================================================================
# event handlers
# ============================================================================

EVENT_HANDLER_NAMES = Literal[
    "go",
    "exit",
]

MAIN_PAGE_EVENT_HANDLERS: dict[EVENT_HANDLER_NAMES, Callable[[], None]] = {}

for event_name in get_args(EVENT_HANDLER_NAMES):
    MAIN_PAGE_EVENT_HANDLERS[event_name] = lambda : print(f"'selected. {event_name}")


def set_main_page_event_handler(name: EVENT_HANDLER_NAMES, handler: Callable[[], None]):
    MAIN_PAGE_EVENT_HANDLERS[name] = handler

# ============================================================================
# class: AllocationManagerTk
# ============================================================================
class AllocationManagerTk(MainPageBaseTk):

    # ----------------------------------------------------------------------------------------------------------------
    # constructor
    # ----------------------------------------------------------------------------------------------------------------
    def __init__(self, title: str, preferences: Preferences = None, bin_dir: str = "."):
        """
        create object - starts with welcome page displayed
        :param title: Sir, Ma'am, Her Royal Highness
        :param preferences: what d'ya want, d'ya really really want?
        :param bin_dir: the directory that the logo for this app is located
        """
        self.preferences: Preferences = preferences
        self.chosen_file_labels: dict[SemesterType, Optional[tk.Label]] = {s: None for s in SemesterType}

        super().__init__(title, preferences)
        MainPageBaseTk.bin_dir = bin_dir

        # select previous files by default
        self.selected_files: dict[SemesterType, tk.StringVar] = {}
        for semester in SemesterType:
            if preferences is not None:
                self.preferences.semester(semester.name)
                self.selected_files[semester] =  tk.StringVar(value=str(self.preferences.previous_file()))
            else:
                self.selected_files[semester] =  tk.StringVar(value=None)

        # create icon for program
        ico = Image.open(f"{bin_dir}/allocation_ico.png")
        photo = ImageTk.PhotoImage(ico)
        self.mw.wm_iconphoto(True, photo)

    # ----------------------------------------------------------------------------------------------------------------
    # setting the schedule filename
    # ----------------------------------------------------------------------------------------------------------------
    def schedule_filename(self, semester:SemesterType = SemesterType.any, value:str =""):
        """
        the schedule file is saved as a Tk StringVar, and is trimmed down to a smaller size if necessary
        :param semester:
        :param value:
        :return: the current value of the Tk StringVar
        """

        # trim file name if necessary
        if value is not None and value != "":
            value = os.path.abspath(value)
            if len(value) > MAX_LEN_OF_DISPLAYED_FILENAME:
                value = "... " + value[len(value) - MAX_LEN_OF_DISPLAYED_FILENAME:]
            self.selected_files[semester].set(value)
        else:
            self.selected_files[semester].set(value="Create New")

        # save the filename
        if semester == SemesterType.winter:
            self.status_bar_winter_file_info = value
        else:
            self.status_bar_fall_file_info = value

        # return the value
        return self.selected_files[semester].get()

    # ----------------------------------------------------------------------------------------------------------------
    # override exit event
    # ----------------------------------------------------------------------------------------------------------------
    def _exit_schedule(self, *_):
        self.exit_schedule()

    def exit_schedule(self):
        MAIN_PAGE_EVENT_HANDLERS["exit"]()
        super()._exit_schedule()

    # ----------------------------------------------------------------------------------------------------------------
    # create welcome page
    # ----------------------------------------------------------------------------------------------------------------
    def create_welcome_page(self, valid_semesters:list[SemesterType]):
        """
        Creates the very first page that is shown to the user
        :param valid_semesters:
        """
        option_frame = super().create_welcome_page_base(
            os.path.join(self.bin_dir, 'allocation_logo.png'))

        # divide frame into semesters
        option_frames: dict[SemesterType, tk.LabelFrame] = {}
        for row,semester in enumerate(valid_semesters):
            option_frames[semester] = tk.LabelFrame(option_frame,text=semester.name,font=self.fonts.big)
            option_frames[semester].pack(expand=1, fill='y', side='top')

        for semester in valid_semesters:

            # open previous schedule file buttons
            self.chosen_file_labels[semester]= tk.Label(
                option_frames[semester],
                anchor="e",
                font=self.fonts.big,
                borderwidth=2,
                background=self.colours.ButtonBackground,
                foreground=self.colours.ButtonForeground,
                width=BUTTON_WIDTH,
                height=2,
                textvariable=self.selected_files[semester],
                state='disabled'
            )
            self.chosen_file_labels[semester].pack(side="top", fill="x", expand=1, padx=5, pady=5)

            # open schedule file option
            tk.Button(
                option_frames[semester],
                text=f"Browse",
                font=self.fonts.normal,
                borderwidth=2,
                background=self.colours.ButtonBackground,
                foreground=self.colours.ButtonForeground,
                command=partial(self.select_file,semester),
                width=8,
                height=2
            ).pack(side="top", fill="x", expand=1, padx=5, pady=5)

            # open schedule file option
            tk.Button(
                option_frames[semester],
                text=f"New",
                font=self.fonts.normal,
                borderwidth=2,
                background=self.colours.ButtonBackground,
                foreground=self.colours.ButtonForeground,
                command=partial(self.new_file,semester),
                width=8,
                height=2
            ).pack(side="top", fill="x", expand=1, padx=5, pady=5)

        # go
        tk.Button(option_frame, text="Go", font=self.fonts.big, padx=20,pady=5,
                  command=self.go).pack(expand=1, fill='x', side='top', pady=20)

    def select_file(self, semester):
        file = self.select_file_to_open(f"Open Schedule for {semester}")
        self.selected_files[semester].set(value=file)

    def new_file(self, semester):
        self.selected_files[semester].set(value=str(None))

    def go(self):
        MAIN_PAGE_EVENT_HANDLERS["go"]()


class InvalidSectionNumberForCourseError(Exception):
    """Section Number already exists for specified course"""


class InvalidHoursForSectionError(Exception):
    """Section hours must be greater than zero"""


class InvalidTeacherNameError(Exception):
    """Teacher name cannot be just spaces or an empty string"""


class CouldNotReadFileError(Exception):
    """Could not read the schedule file"""


class CouldNotWriteFileError(Exception):
    """Could not read the schedule file"""



"""
# ============================================================================
# Main entry point to the schedule application
#
# Events triggered by menus_main_menu toolbar and menu events
#       new_menu_event()
#       open_menu_event()
#       save_menu_event()
#       save_as_menu_event()
#       validate()
#       exit_event()
#       print_views(resource_type, canvas)
#       print_registrar()
#       auto_save_set(bool)
#
# Events triggered from the welcome page
#       exit_event()
#       open_menu_event()
#       new_menu_event()
#       open_previous_file_event()
#       semester_change()
#
# Notebook change events
#       update_course_text(frame)
#       update_teacher_text(frame)
#       update_choices_of_resource_views(frame)
#       update_overview(frame)
#       update_edit_courses(frame)
#       update_edit_teachers(frame)
#       update_edit_labs(frame)
#       update_edit_streams(frame)
#
# ============================================================================
"""


from __future__ import annotations

from enum import Enum
from functools import partial
from typing import Optional

from .edit_resources import EditResources
from .edit_courses import EditCourses
from .menus_main_menu_scheduler import set_menu_event_handler, main_menu
from .view import View
from .views_controller import ViewsController
from .notebook_tab_data import NBTabInfo

from ..Utilities import Preferences
from ..export.registrars_office import registrars_output
from ..gui_pages.scheduler_tk import SchedulerTk, set_main_page_event_handler
from ..model import Schedule, ResourceType
from ..model.exceptions import CouldNotReadFileError
from ..gui_generics.read_only_text_tk import ReadOnlyTextTk
from ..export.view_export_canvases import PDFCanvas, LatexCanvas
from ..gui_pages.view_canvas_tk import ViewCanvasTk

class CanvasType(Enum):
    latex = 1
    pdf = 2

# =====================================================================================
# Scheduler
# =====================================================================================

DIRECTORY = str


class Scheduler:
    """
    This is the main entry point for the Scheduler Program
    """

    NB_overview_course = "by Course"
    NB_overview_teacher = "by Teacher"
    NB_schedule = "Schedules"
    NB_overview = "Overview"
    NB_course = "Courses"
    NB_teacher = "Teachers"
    NB_lab = "Labs"
    NB_stream = "Streams"

    def __init__(self, bin_dir: DIRECTORY, gui: Optional[SchedulerTk] = None):
        """
        :param bin_dir: where this executable file exists
        :param gui: creates gui if not already defined
        """

        self.view_controller: Optional[ ViewsController] = None
        self.preferences: Preferences = Preferences()
        self.schedule: Optional[Schedule] = None
        self._dirty_flag = False
        self.current_tab: Optional[str] = None

        # gui is optional so that we can test the presenter more readily
        if gui:
            self.gui = gui
        else:
            self.gui: SchedulerTk = SchedulerTk('Scheduler', self.preferences, bin_dir)

        self._schedule_filename = ""
        self._previous_filename = self.preferences.previous_file()

        # --------------------------------------------------------------------
        # required notebook pages
        # --------------------------------------------------------------------
        by_course_page = NBTabInfo(label=self.NB_overview_course, name=self.NB_overview_course)
        by_teacher_page = NBTabInfo(label=self.NB_overview_teacher, name=self.NB_overview_teacher)

        self._required_tabs: list[NBTabInfo] = [
            NBTabInfo(label=self.NB_schedule, name=self.NB_schedule, is_default_page=True),
            NBTabInfo(label=self.NB_overview, name=self.NB_overview, subpages=[by_course_page, by_teacher_page]),
            NBTabInfo(label=self.NB_course, name=self.NB_course),
            NBTabInfo(label=self.NB_teacher, name=self.NB_teacher),
            NBTabInfo(label=self.NB_lab, name=self.NB_lab),
            NBTabInfo(label = self.NB_stream, name = self.NB_stream)
        ]

        # --------------------------------------------------------------------
        # create the Menu and Toolbars
        # --------------------------------------------------------------------
        set_menu_event_handler("file_new", self.new_menu_event)
        set_menu_event_handler("file_open", self.open_menu_event)
        set_menu_event_handler("file_save", self.save_menu_event)
        set_menu_event_handler("file_save_as", self.save_as_menu_event)
        set_menu_event_handler("validate", self.validate)
        set_menu_event_handler("file_exit", self.menu_exit_event)
        set_main_page_event_handler("file_exit", self.exit_event)
        set_main_page_event_handler("file_open", self.open_menu_event)
        set_main_page_event_handler("file_new", self.new_menu_event)
        set_main_page_event_handler("file_open_previous", self.open_previous_file_event)
        set_main_page_event_handler("semester_change", self.semester_change)
        set_menu_event_handler("print_pdf_teacher", partial(self.print_views, ResourceType.teacher, CanvasType.pdf))
        set_menu_event_handler("print_pdf_lab", partial(self.print_views, ResourceType.lab, CanvasType.pdf))
        set_menu_event_handler("print_pdf_streams", partial(self.print_views, ResourceType.stream, CanvasType.pdf))
        set_menu_event_handler("print_latex_teacher", partial(self.print_views, ResourceType.teacher, CanvasType.latex))
        set_menu_event_handler("print_latex_lab", partial(self.print_views, ResourceType.lab, CanvasType.latex))
        set_menu_event_handler("print_latex_streams", partial(self.print_views, ResourceType.stream, CanvasType.latex))
        set_menu_event_handler("print_registrar", self.print_registrar)

        self.gui.toggle_auto_save = self.auto_save_set

        # TODO
        set_menu_event_handler("print_text", self.menu_ignore)

        (self._toolbar_buttons, self._button_properties, self._menu) = main_menu()

        # --------------------------------------------------------------------
        # create the Gui Main Window
        # --------------------------------------------------------------------
        self.gui.create_menu_and_toolbars(self._toolbar_buttons, self._button_properties, self._menu)
        self.gui.create_welcome_page(self.preferences.semester())
        self.gui.create_status_bar()
        self.gui.notebook_tab_changed_handler = self.notebook_tab_has_changed
        self.set_dirty_indicator()

        self.gui.start_event_loop()

    # ============================================================================================
    #  Properties - filenames
    # ============================================================================================
    @property
    def schedule_filename(self):
        """The filename associated with this file"""
        return self._schedule_filename

    @schedule_filename.setter
    def schedule_filename(self, value: Optional[str]):
        self._schedule_filename = value
        self.gui.schedule_filename = value
        if value is not None and value != "":
            self.previous_filename = value

    @property
    def previous_filename(self):
        """what was the last opened file"""
        return self._previous_filename

    @previous_filename.setter
    def previous_filename(self, value):
        self.preferences.previous_file(value)
        self._previous_filename = value
        self.gui.previous_file = value
        self.preferences.save()

    @property
    def auto_save_text(self):
        if self.preferences.auto_save():
            return "AUTO SAVE ON"
        else:
            return ""

    # ============================================================================================
    # Properties - is data changed (dirty)
    # ============================================================================================
    @property
    def dirty_flag(self) -> bool:
        """is the data different than what was saved on disk?"""
        return self._dirty_flag

    @dirty_flag.setter
    def dirty_flag(self, value):
        self.set_dirty_method(value)

    # ============================================================================================
    # Event handler, auto save setting changed
    # ============================================================================================
    def set_dirty_indicator(self):
        if self.gui and self.dirty_flag:
            self.gui.dirty_text = "UNSAVED"
        else:
            self.gui.dirty_text = self.auto_save_text

    def auto_save_set(self, value):
        self.preferences.auto_save(value)
        self.set_dirty_indicator()
        self.preferences.save()

    # ============================================================================================
    # Event handlers - file open/close/save/new
    # ============================================================================================
    def new_menu_event(self):
        """create a new file"""
        self.schedule = Schedule()
        self.schedule_filename = ""
        self.dirty_flag = True
        self.refresh_for_newly_opened_file()

    def open_menu_event(self):
        """open a file"""
        filename = self.gui.select_file_to_open(f"Open Schedule")
        self._open_file(filename)

    def open_previous_file_event(self):
        """open previously opened file"""
        filename = self.preferences.previous_file()
        self._open_file(filename)

    def _open_file(self, filename: str):
        """generic open file method"""
        if filename:
            try:
                schedule = Schedule(filename)
                self.schedule = schedule
                self.schedule_filename = filename
                self.dirty_flag = False
                self.refresh_for_newly_opened_file()

            except CouldNotReadFileError as e:
                self.gui.show_custom_message("Read File Error", str(e))

    def save_menu_event(self):
        """save file"""
        self._save_schedule(self.schedule_filename)

    def save_as_menu_event(self):
        """save as file"""
        self._save_schedule(None)

    def _save_schedule(self, filename: Optional[str]):
        """generic save file method"""

        if self.schedule is None:
            self.gui.show_error("Save Schedule", "There is no schedule to save!")
            return
        if filename is None or filename == "":
            filename = self.gui.select_file_to_save()

        if filename is not None and filename != "":
            self.schedule.write_file(filename)
            self.schedule_filename = filename
            self.dirty_flag = False

    # ============================================================================================
    # Event handlers - semester change
    # ============================================================================================
    def semester_change(self, *_):
        """semester has been changed"""
        self.preferences.semester(self.gui.current_semester)
        self.previous_filename = self.preferences.previous_file()
        self.preferences.save()

    def refresh_for_newly_opened_file(self):
        pass
        if self.view_controller is not None:
            self.view_controller.kill_all_views()
        self.view_controller = None
        self.gui.create_standard_page(self._required_tabs, reset=True)

    # ============================================================================================
    # Event handlers - exit
    # ============================================================================================

    def exit_event(self):
        """program is exiting"""
        if self.dirty_flag:
            ans = self.gui.ask_yes_no("File", "Save File?")
            if ans:
                self.save_menu_event()

    def menu_exit_event(self):
        self.gui.exit_schedule()

    def menu_ignore(self):
        # print("still a To Do!!")
        pass

    # ==================================================================
    # notebook tab has changed
    # ==================================================================
    def notebook_tab_has_changed(self, name: str, frame):
        """
        the standard page notebook tab has been change
        :param name: name of the tab
        :param frame: container where the gui is stored
        """
        self.current_tab = name
        if name == self.NB_overview_course:
            self.update_course_text(frame)
        elif name == self.NB_overview_teacher:
            self.update_teacher_text(frame)
        elif name == self.NB_schedule:
            self.update_choice_colours(frame)
        elif name == self.NB_overview:
            self.update_overview(frame)
        elif name == self.NB_course:
            self.update_edit_courses(frame)
        elif name == self.NB_teacher:
            self.update_edit_teachers(frame)
        elif name == self.NB_lab:
            self.update_edit_labs(frame)
        elif name == self.NB_stream:
            self.update_edit_streams(frame)

    # ==================================================================
    # update choice colours
    # ==================================================================
    def update_choice_colours(self, frame):
        """if not already created, create Views Controller, update colours"""
        if self.view_controller is None:
            self.view_controller = ViewsController(self.set_dirty_method, frame, self.schedule)
        self.view_controller.refresh()

    # ==================================================================
    # update_overview
    # A text representation of the schedules
    # ==================================================================
    def update_overview(self, frame):
        """not required to do anything"""
        pass

    def update_course_text(self, frame):
        """update text describing all the courses"""
        text: list[str] = list()
        if self.schedule is None:
            text.append('There is no schedule, please open one')
        else:
            if not self.schedule.courses:
                text.append('No courses defined in this schedule')
            else:
                for c in self.schedule.courses():
                    text.append(str(c))
        data_entry = ReadOnlyTextTk(frame, text)
        data_entry.write(text)

    def update_teacher_text(self, frame):
        """udpate the text describing all the teacher duties"""
        text = []
        if not self.schedule.teachers:
            text.append('No teachers defined in this schedule')
        else:
            for t in self.schedule.teachers():
                text.append(self.schedule.teacher_details(t))
        data_entry = ReadOnlyTextTk(frame, text)
        data_entry.write(text)

    # ==================================================================
    # update_edit_teachers
    # ==================================================================
    def update_edit_teachers(self, frame):
        """A page where teacher can be added/modified or deleted"""
        data_entry = EditResources(self.set_dirty_method, frame, ResourceType.teacher, self.schedule, self.preferences)
        data_entry.schedule = self.schedule
        data_entry.refresh()

    # ==================================================================
    # update_edit_streams
    # ==================================================================
    def update_edit_streams(self, frame):
        """A page where stream_ids can be added/modified or deleted"""
        data_entry = EditResources(self.set_dirty_method, frame, ResourceType.stream, self.schedule, self.preferences)
        data_entry.schedule = self.schedule
        data_entry.refresh()

    # ==================================================================
    # update_edit_labs
    # ==================================================================
    def update_edit_labs(self, frame):
        """A page where lab_ids can be added/modified or deleted"""
        data_entry = EditResources(self.set_dirty_method, frame, ResourceType.lab, self.schedule, self.preferences)
        data_entry.schedule = self.schedule
        data_entry.refresh()

    # ==================================================================
    # draw_edit_courses
    # ==================================================================
    def update_edit_courses(self, frame):
        """A page where courses can be added/modified or deleted"""
        data_entry = EditCourses(self.set_dirty_method, frame, self.schedule)
        data_entry.schedule = self.schedule
        data_entry.refresh()


    # ==================================================================
    # print_views
    # ==================================================================
    def print_views(self, resource_type, canvas_type: CanvasType):
        """
        print the schedule 'views'
        :param resource_type:
        :param canvas_type: pdf or latex
        :return:
        """

        # no schedule yet
        if not self.schedule:
            self.gui.show_error("Export", "Cannot export - There is no schedule")

        # should not print if the schedule is not saved
        if self.dirty_flag:
            self.gui.show_message(
                title="Unsaved Changes",
                msg="There are unsaved changes",
                detail="You cannot print schedules if the schedule is not saved"
                )
            return

        # gather info
        resources = ()
        match resource_type:
            case ResourceType.teacher:
                resources = self.schedule.teachers()
            case ResourceType.lab:
                resources = self.schedule.labs()
            case ResourceType.stream:
                resources = self.schedule.streams()

        # loop over each resource and create file
        save_dir = self.preferences.current_dir() or self.preferences.home_directory()
        for resource in resources:

            # pdf or latex
            if canvas_type == CanvasType.pdf:
                cn = PDFCanvas(title=str(resource), schedule_name=self.schedule.filename,
                               directory=save_dir)
                vc = ViewCanvasTk(cn, 0.8)
            else:
                cn = LatexCanvas(title=str(resource), schedule_name=self.schedule.filename,
                               directory=save_dir)
                vc = cn

            # get all the blocks
            blocks = ()
            match resource_type:
                case ResourceType.teacher:
                    blocks = self.schedule.get_blocks_for_teacher(resource)
                case ResourceType.lab:
                    blocks = self.schedule.get_blocks_in_lab(resource)
                case ResourceType.stream:
                    blocks = self.schedule.get_blocks_for_stream(resource)

            # draw the blocks
            for block in blocks:
                text = View.get_block_text(block, scale=1, resource_type=ResourceType.teacher)

                day, start_time, duration = View._block_to_floats(block)

                vc.draw_block(resource_type=ResourceType.teacher,
                              day=day, start_time=start_time, duration=duration,
                              text=text,
                              gui_tag="",
                              movable=True,
                              )
            # write to file
            cn.save()

    # ==================================================================
    # print the output needed for the registrar
    # ==================================================================
    def print_registrar(self):
        if self.schedule is None:
            self.gui.show_error("Save Schedule", "There is no schedule to write!")
            return
        filename = self.gui.select_file_to_save()
        try:
            registrars_output(self.schedule, filename)
        except Exception as e:
            self.gui.show_message(title="Registrar's Output", msg="Could not save file",
                                  detail=f"Error: {e}")


    # ==================================================================
    # validate
    # ==================================================================
    def validate(self):
        """check if the schedule is valid, show result to user"""
        msg = self.schedule.validate()
        if len(msg) != 0:
            self.gui.show_text(title="Validate", text=msg)
        else:
            self.gui.show_message(title="Validate", msg="Everything is ok!")


    # ==================================================================
    # schedule has been modified, update gui as required
    # ==================================================================
    def set_dirty_method(self, value: Optional[bool] = None) -> bool:

        # if value is true, update all the open views
        if value and self.current_tab != self.NB_schedule and self.view_controller is not None:
            self.view_controller.redraw_all()

        # if value is true, and autosave is on, save the file
        if value and self.preferences.auto_save():
            self._save_schedule(self.schedule_filename)
            value = False

        if value is not None:
            self._dirty_flag = value

        self.set_dirty_indicator()
        return self.dirty_flag




import os
import tkinter as tk

from src.scheduling_and_allocation.model import Schedule
from src.scheduling_and_allocation.modified_tk import set_default_fonts_and_colours
from src.scheduling_and_allocation.presenter.allocation_editor import AllocationEditor

bin_dir: str = os.path.dirname(os.path.realpath(__file__))
mw = tk.Tk()
mw.geometry("800x400")

_ = set_default_fonts_and_colours(mw)
file = f"{bin_dir}/../unit_tests_presenter/data_fall.csv"
s = Schedule(file)

f = tk.Frame(mw, background="pink")
f.pack(expand=1, fill="both")
def set_dirty_flag(flag=False): ...

ae = AllocationEditor(set_dirty_flag, frame =f, schedule = s)

tk.mainloop()




"""
The over-arching class that contains ALL the information for scheduler and allocation programs
"""
from __future__ import annotations

from os import path
from typing import Optional

from .exceptions import CouldNotWriteFileError, CouldNotReadFileError

from .teacher import Teacher
from .block import Block
from .course import Course
from .lab import Lab
from .stream import Stream
from .section import Section
from .enums import ConflictType
from .conflicts import (set_block_conflicts, set_lunch_break_conflicts,
                        set_number_of_days_conflict, set_availability_hours_conflict, block_conflicts_time,
                        has_lunch_break_conflict, has_number_of_days_conflict, has_availability_hours_conflict)
from .enums import ResourceType, SemesterType
from .serializor import CSVSerializor as Serializor

DEFAULT_COURSE_HOURS = 3



def get_resource_type(obj: Teacher | Lab | Stream) -> ResourceType | None:
    """Returns the resource_type of the ResourceType object"""
    for v_type in ResourceType:
        my_class = eval(f"{v_type.name}")
        if isinstance(obj, my_class):
            return v_type
    return None

# ============================================================================
# Schedule
# ============================================================================
class Schedule:
    """ Provides the top level class for all schedule objects. """

    # ------------------------------------------------------------------------
    # Initialize
    # ------------------------------------------------------------------------
    def __init__(self, file: Optional[str] = None):
        """ Creates an instance of the Schedule class. """
        self._teachers: dict[str, Teacher] = dict()
        self._streams: dict[str, Stream] = dict()
        self._labs: dict[str, Lab] = dict()
        self._courses: dict[str, Course] = dict()
        self.filename = ""

        if file is not None:
            self.read_file(file)

    # ------------------------------------------------------------------------
    # read file
    # ------------------------------------------------------------------------
    def read_file(self, file):
        """read a csv file containing the schedule info"""
        try:
            Serializor.read_file(self, file)
        except Exception as e:
            msg = (f"Could not read\n {file}\n\n"
                   f"Line {Serializor.last_line_number_read}: {Serializor.last_line_read}\n\n"
                   f"Error Message: {e}")
            raise CouldNotReadFileError(msg)
        self.filename = path.basename(file)

    # ------------------------------------------------------------------------
    # write file
    # ------------------------------------------------------------------------
    def write_file(self, file):
        """write to a csv file all the info about the schedule"""
        try:
            Serializor.write(self, file)
        except Exception as e:
            raise CouldNotWriteFileError(f"Could not read {file}, {e}")
        self.filename = path.basename(file)

    # ------------------------------------------------------------------------
    # add/update course
    # ------------------------------------------------------------------------
    def add_update_course(self, number: str, name: str = "new Course",
                          semester: SemesterType = SemesterType.any, hours:float = DEFAULT_COURSE_HOURS,
                          needs_allocation: bool = True) -> Course:
        """Creates or updates a Course object with the unique identifier 'teacher_id'
        :param number: Unique identifier for the course
        :param name: Name of the course
        :param semester: Which semester is this course in
        :param hours: How many hours per week are allocated for this course
        :param needs_allocation: Will give allocation to any teacher who teaches this course [True]
        """
        original_course: Course = self.get_course_by_number(number)
        if original_course is None:
            course: Course = Course(number, name, semester, hours, needs_allocation)
            self._courses[course.number] = course
            return course
        else:
            original_course.semester = semester
            original_course.name = name
            original_course.needs_allocation = needs_allocation
            return original_course

    # ------------------------------------------------------------------------
    # add/update stream
    # ------------------------------------------------------------------------
    def add_update_stream(self, number: str, description: str = "") -> Stream:
        """Creates or updates a Stream object with the unique identifier 'teacher_id'
        :param number: Unique identifier for the stream
        :param description: (optional)
        """
        original_stream = self.get_stream_by_number(number)
        if original_stream is None:
            stream = Stream(number, description)
            self._streams[stream.number] = stream
            return stream
        else:
            original_stream.description = description
            return original_stream

    # ------------------------------------------------------------------------
    # add/update lab
    # ------------------------------------------------------------------------
    def add_update_lab(self, number: str, description: str = '') -> Lab:
        """Creates or updates a Lab object with the unique identifier 'teacher_id'
        :param number: Unique identifier for the lab
        :param description: (optional)
        """
        original_lab = self.get_lab_by_number(number)
        if original_lab is None:
            lab: Lab = Lab(number, description)
            self._labs[lab.number] = lab
            return lab
        else:
            original_lab.description = description
            return original_lab

    # ------------------------------------------------------------------------
    # add/update teacher
    # ------------------------------------------------------------------------
    def add_update_teacher(self, firstname: str, lastname: str, department: str = "", release:float = 0,
                           teacher_id: str = None) -> Teacher:
        """Creates or updates a teacher object with the unique identifier 'teacher_id'
        :param release:
        :param firstname:
        :param lastname:
        :param release: (optional) - how much release does the teach have? fractional FTE
        :param department: (optional)
        :param teacher_id: (optional) - if not specified, a unique id will be created
        """
        original_teacher = self.get_teacher_by_number(teacher_id) if teacher_id is not None else None
        if original_teacher is None:
            teacher = Teacher(firstname, lastname, department, release=release)
            self._teachers[teacher.number] = teacher
            return teacher
        else:
            original_teacher.firstname = firstname
            original_teacher.lastname = lastname
            original_teacher.department = department
            original_teacher.release = release
            return original_teacher

    # ------------------------------------------------------------------------
    # returning collections as a tuple (to prevent user from modifying the collection)
    # ------------------------------------------------------------------------
    def courses(self) -> tuple[Course, ...]:
        """sorted list of courses (read only)"""
        return tuple(sorted(self._courses.values()))

    def labs(self) -> tuple[Lab, ...]:
        """sorted list of labs (read only)"""
        return tuple(sorted(self._labs.values()))

    def streams(self) -> tuple[Stream, ...]:
        """sorted list of streams (read only)"""
        return tuple(sorted(self._streams.values()))

    def teachers(self) -> tuple[Teacher, ...]:
        """sorted list of teachers (read only)"""
        return tuple(sorted(self._teachers.values()))

    def courses_with_allocation(self)-> tuple[Course, ...]:
        return tuple(sorted((c for c in self.courses() if c.needs_allocation)))


    # ------------------------------------------------------------------------
    # finding object to collection by number, id, etc
    # ------------------------------------------------------------------------

    def get_course_by_number(self, number: str) -> Optional[Course]:
        """Returns the Course which matches this Course number, if it exists."""
        return self._courses.get(number)

    def get_lab_by_number(self, number: str) -> Optional[Lab]:
        """Returns the Lab which matches this Lab number, if it exists."""
        return self._labs.get(number)

    def get_stream_by_number(self, number: str) -> Optional[Stream]:
        """Returns the Stream which matches this Stream number, if it exists."""
        return self._streams.get(number)

    def get_teacher_by_name(self, firstname: str, lastname: str) -> Optional[Teacher]:
        """Returns the Teacher which matches this name, if it exists."""
        found = [t for t in self._teachers.values()
                 if t.firstname.lower() == firstname.lower() and
                 t.lastname.lower() == lastname.lower()]
        return found[0] if found else None

    def get_teacher_by_number(self, teacher_id: str) -> Optional[Teacher]:
        """Returns the teacher which matches this Section number, if it exists."""
        teacher_id = str(teacher_id)
        return self._teachers.get(teacher_id)

    def get_view_type_obj_by_id(self, view_type: ResourceType, obj_id: str) -> Optional[Teacher | Stream | Lab]:
        """Returns the Section which matches this Section number, if it exists."""
        if view_type == ResourceType.teacher:
            return self.get_teacher_by_number(obj_id)
        elif view_type == ResourceType.stream:
            return self.get_stream_by_number(obj_id)
        elif view_type == ResourceType.lab:
            return self.get_lab_by_number(obj_id)
        return None

    # ========================================================================
    # removing objects from the collections
    # - this may require detaching some dependencies
    # ========================================================================

    def remove_course(self, course: Course):
        """Removes Course from the collection of courses"""
        self._courses.pop(course.number, None)

    def remove_teacher(self, teacher: Teacher):
        """Removes Teacher from all scheduled courses and from the collection of teachers"""
        for c in self.courses():
            c.remove_teacher(teacher)
        self._teachers.pop(teacher.number, None)

    def remove_lab(self, lab: Lab):
        """Removes Lab from all blocks where it is used, and removes from collection of labs"""
        for b in self.blocks():
            b.remove_lab(lab)
        self._labs.pop(lab.number, None)

    def remove_stream(self, stream: Stream):
        """Removes Stream from all sections where it is used and removes from collection of streams"""
        for s in self.sections():
            s.remove_stream(stream)
        self._streams.pop(stream.number, None)

    # ========================================================================
    # filtered collections
    # ========================================================================

    def get_teachers_assigned_to_any_course(self) -> tuple[Teacher, ...]:
        """Returns a tuple of all the Teacher objects with assigned courses"""
        teachers: set[Teacher] = set()
        for c in self.courses():
            for s in c.sections():
                teachers.update(set(s.teachers()))
        return tuple(sorted(teachers))

    def get_streams_assigned_to_any_course(self) -> tuple[Stream, ...]:
        """Returns a tuple of all the Stream objects that have been assigned to any section in a course"""
        streams: set[Stream] = set()
        for c in self.courses():
            for s in c.sections():
                streams.update(set(s.streams()))
        return tuple(sorted(streams))

    def get_labs_assigned_to_any_course(self) -> tuple[Lab, ...]:
        """Returns a tuple of all the Lab objects that have been assigned to any block in a course"""
        labs: set[Lab] = set()
        for c in self.courses():
            for s in c.sections():
                labs.update(set(s.labs()))
        return tuple(sorted(labs))

    def get_courses_for_teacher(self, teacher: Teacher) -> tuple[Course, ...]:
        """Get all the courses that has this teacher assigned to it"""
        courses: set[Course] = set([c for c in self.courses() if c.has_allocated_teacher(teacher)])
        return tuple(sorted(courses))

    # ========================================================================
    # Course innards
    # ========================================================================
    def sections(self) -> tuple[Section, ...]:
        """ Returns a tuple of all the schedule's Section objects in this Schedule"""
        sections: set[Section] = set()
        for c in self.courses():
            sections.update(set(c.sections()))
        return tuple(sections)

    def blocks(self) -> tuple[Block, ...]:
        """ Returns a tuple of all the schedule's Block objects in this Schedule"""
        blocks: set[Block] = set()
        for s in self.sections():
            blocks.update(set(s.blocks()))
        return tuple(blocks)

    # ========================================================================
    # filtered Course innards
    # ========================================================================
    def get_sections_for_teacher(self, teacher: Teacher) -> tuple[Section, ...]:
        """Returns a tuple of Sections that the given Teacher teaches"""
        sections: set[Section] = set([s for s in self.sections() if teacher in s.teachers()])
        return tuple(sections)

    def get_sections_for_stream(self, stream: Stream) -> tuple[Section, ...]:
        """Returns a tuple of Sections assigned to the given Stream"""
        streams: set[Section] = set([s for s in self.sections() if s.has_stream(stream)])
        return tuple(streams)

    def get_blocks_for_teacher(self, teacher: Teacher) -> tuple[Block, ...]:
        """Returns a tuple of Blocks that the given Teacher teaches"""
        blocks: set[Block] = set([b for b in self.blocks() if b.has_teacher(teacher)])
        return tuple(blocks)

    def get_blocks_in_lab(self, lab: Lab) -> tuple[Block, ...]:
        """Returns a tuple of Blocks using the given Lab"""
        blocks: set[Block] = set([b for b in self.blocks() if b.has_lab(lab)])
        return tuple(blocks)

    def get_blocks_for_stream(self, stream: Stream) -> tuple[Block, ...]:
        """Returns a tuple of blocks in a given stream"""
        blocks: set[Block] = set()
        for s in self.get_sections_for_stream(stream):
            blocks.update(set(s.blocks()))
        return tuple(blocks)

    def get_blocks_for_obj(self, obj: Teacher | Lab | Stream) -> tuple[Block, ...]:
        """ Returns a tuple of blocks associated with the specified ResourceType object"""
        if isinstance(obj, Teacher):
            return self.get_blocks_for_teacher(obj)
        if isinstance(obj, Lab):
            return self.get_blocks_in_lab(obj)
        if isinstance(obj, Stream):
            return self.get_blocks_for_stream(obj)
        return tuple()

    # --------------------------------------------------------
    # teacher_details
    # --------------------------------------------------------

    def teacher_details(self, teacher: Teacher) -> str:
        """
        Prints a schedule for a specific teacher
        Parameter teacher -> The teacher whose schedule to print
        """
        head = "=" * 50
        text = f"\n\n{head}\n{teacher}\n{head}\n"

        c: Course
        for c in sorted(self.get_courses_for_teacher(teacher), key=lambda a: a.number.lower()):

            text += f"\n{c.number} {c.name}\n"
            text += "-" * 80

            # sections
            for s in sorted(c.sections(), key=lambda a: a.number):
                if s.has_teacher(teacher):
                    text += f"\n\t{s}\n\t" + "- " * 25 + "\n"

                    # blocks
                    for b in s.blocks():
                        if b.has_teacher(teacher):
                            text += f"\t{b} "
                            text += ", ".join(str(lab) for lab in b.labs()) + "\n"
        return text

    # --------------------------------------------------------
    # clear_all_from_course
    # --------------------------------------------------------
    def clear_all_from_course(self, course: Course):
        """
        Removes all teacher_ids, lab_ids, and stream_ids from course
        - Parameter course -> The course to be cleared.
        """
        for section in self.sections():
            if section.course is course:
                section.remove_all_teachers()
                section.remove_all_streams()
                section.remove_all_labs()

    # --------------------------------------------------------
    # Calculate Conflicts
    # --------------------------------------------------------
    def calculate_conflicts(self):
        """Reviews the schedule, and creates a list of Conflict objects as necessary"""

        # reset all blocks conflicted_number tags
        for b in self.blocks():
            b.conflict = ConflictType.NONE

        # set conflicts for each block, as required
        for teacher in self.get_teachers_assigned_to_any_course():
            set_block_conflicts(self.get_blocks_for_teacher(teacher), ConflictType.TIME_TEACHER)
        for stream in self.get_streams_assigned_to_any_course():
            set_block_conflicts(self.get_blocks_for_stream(stream), ConflictType.TIME_STREAM)
        for lab in self.get_labs_assigned_to_any_course():
            set_block_conflicts(self.get_blocks_in_lab(lab), ConflictType.TIME_LAB)

        # ---------------------------------------------------------
        # for each teacher teacher
        # ---------------------------------------------------------
        for teacher in self.get_teachers_assigned_to_any_course():
            set_lunch_break_conflicts(self.get_blocks_for_teacher(teacher))
            set_availability_hours_conflict(self.get_blocks_for_teacher(teacher))
            if teacher.release == 0:
                set_number_of_days_conflict(self.get_blocks_for_teacher(teacher))

    # --------------------------------------------------------
    # get conflict for a specific resource
    # --------------------------------------------------------
    def resource_conflict(self, resource: Teacher|Stream|Lab) -> ConflictType:
        """Calculate the overall conflict for a given resource"""
        block_conflict: ConflictType = ConflictType.NONE

        for block in self.get_blocks_for_obj(resource):
            block_conflict = block.conflict | block_conflict

        resource_conflict: ConflictType = ConflictType.NONE
        if (block_conflict.is_time_teacher() | block_conflict.is_time_lab()
                or block_conflict.is_time_stream()) | block_conflict.is_time():
            resource_conflict = ConflictType.TIME

        match resource.resource_type:
            case ResourceType.teacher:
                if block_conflicts_time(self.get_blocks_for_teacher(resource)):
                    resource_conflict = resource_conflict | ConflictType.TIME_TEACHER

                if has_lunch_break_conflict(self.get_blocks_for_teacher(resource)):
                    resource_conflict = resource_conflict | ConflictType.LUNCH

                if has_availability_hours_conflict(self.get_blocks_for_teacher(resource)):
                    resource_conflict = resource_conflict | ConflictType.AVAILABILITY

                if resource.release == 0:
                    if has_number_of_days_conflict(self.get_blocks_for_teacher(resource)):
                        resource_conflict = resource_conflict | ConflictType.MINIMUM_DAYS


            case ResourceType.lab:
                if block_conflicts_time(self.get_blocks_in_lab(resource)):
                    resource_conflict = resource_conflict | ConflictType.TIME_LAB

            case ResourceType.stream:
                if block_conflicts_time(self.get_blocks_for_stream(resource)):
                    resource_conflict = resource_conflict | ConflictType.TIME_STREAM

        return resource_conflict

    # --------------------------------------------------------
    # validate that the schedule is good
    # --------------------------------------------------------
    def validate(self) -> list[str]:
        """Validates that the schedule makes sense"""
        msg = []

        # each course that has allocation has at least one teacher assigned
        for course in self.courses():
            course_flag = False
            block_flag = False
            section_flag = False
            if course.needs_allocation:
                if len(course.teachers()) == 0:
                    msg.append(f"\nERROR: Course {course.number} has no teachers")
                    course_flag = True

            # each block should have at least one teacher
            for block in course.blocks():
                if len(block.teachers()) == 0:
                    if not course_flag and not block_flag:
                        msg.append("")
                    block_flag = True
                    msg.append(f"ERROR: {block.section.course.number}, {block.section.number} {block} has no assigned teachers")

            # each block should have at least one lab
            for block in course.blocks():
                if len(block.labs()) == 0:
                    if not course_flag and not block_flag:
                        msg.append("")
                    block_flag = True
                    msg.append(f"WARNING: {block.section.course.number}, {block.section.number} {block} has no assigned labs")

            # each course that has blocks/sections, block time should equal class time
            for section in course.sections():
                duration = sum((b.duration for b in section.blocks()))
                if duration != course.hours_per_week:
                    if not course_flag and not block_flag and not section_flag:
                        msg.append("")
                    msg.append(
                        f"ERROR: {course.number}: {section} - course hours ({course.hours_per_week}) "
                        f"does not match assigned class time ({duration})")

        return msg

    # *** These functions were created so that the gui dialog boxes call model, instead of having
    #     model logic in the presenter

    # -------------------------------------------------------------------------------------------------------------
    # update a section with teachers and blocks and labs and streams
    # -------------------------------------------------------------------------------------------------------------
    def update_section(self, section: Section, name: str, teachers: list[Teacher], labs: list[Lab], streams: list[Stream],
                        blocks: list[tuple[float, float, float]]):
        """
        update a section with teachers and blocks and labs and streams
        :param section:
        :param name: description of the section
        :param teachers: a list of teachers to add to the section
        :param labs: a list of labs to add to the section
        :param streams: a list of streams to add to the section
        :param blocks: a list of tuples (day, start, duration), where day is an integer
        """
        section.name = name
        section.clear()

        for b in blocks:
            section.add_block(*b)
        for t in teachers:
            section.add_teacher(t)
        for l in labs:
            section.add_lab(l)
        for s in streams:
            section.add_stream(s)

    # -------------------------------------------------------------------------------------------------------------
    # Section - add/edit dialog apply
    # -------------------------------------------------------------------------------------------------------------
    def add_sections(self, course, number: int, blocks):
        """
        create a new section
        :param course: the course that this section belongs to
        :param number: the number of sections to add
        :param blocks: a list of tuples (day, start, duration), where day is an integer
        """
        for i in range(number):
            section = course.add_section()
            self.update_section(section, section.name, [], [], [], blocks)

    # -------------------------------------------------------------------------------------------------------------
    # Blocks - add dialog apply
    # -------------------------------------------------------------------------------------------------------------
    def add_blocks(self, section: Section, number: int, hours, teachers, labs):
        for i in range(number):
            block = section.add_block(duration=hours)
            for t in teachers:
                block.add_teacher(t)
            for l in labs:
                block.add_lab(l)

    # -------------------------------------------------------------------------------------------------------------
    # Blocks - edit dialog apply
    # -------------------------------------------------------------------------------------------------------------
    def edit_block(self, block, _, hours, teachers, labs):
        block.remove_all_labs()
        block.remove_all_teachers()
        block.duration = hours
        for t in teachers:
            block.add_teacher(t)
        for l in labs:
            block.add_lab(l)

    # -------------------------------------------------------------------------------------------------------------
    # Course - add/edit dialog apply changes
    # -------------------------------------------------------------------------------------------------------------
    def add_edit_course(self, course_number: str, course_name: str, hours_per_week: float,
                                      allocation: bool, num_sections: int, teachers: list[Teacher],
                                      labs: list[Lab], blocks: list[tuple[float, float, float]]):
        """
        Apply changes from course add/edit dialog box
        :param course_number: a unique number
        :param course_name:
        :param hours_per_week:
        :param allocation: do you need to assign a teacher
        :param num_sections:
        :param teachers: teachers to assign to all sections of this course
        :param labs: labs to assign to all sections of this course
        :param blocks: blocks to assign to all sections of this course
        """
        if course_number not in (c.number for c in self.courses()):
            course = self.add_update_course(number=course_number)
        else:
            course = self.get_course_by_number(course_number)

        course.name = course_name
        course.hours_per_week = hours_per_week
        course.needs_allocation = allocation
        course.remove_all_sections()
        for _ in range(num_sections):
            section = course.add_section()
            self.update_section(section, "", teachers, labs, [], blocks)




import re, platform
import os

dir_path = os.path.dirname(os.path.realpath(__file__))
#from schedule.Presentation import dirty_flags


def get_image_dir():
    """Gets the image directory"""
    return os.path.join(dir_path, 'Images')


def get_logo(bin_dir):
    return os.path.join(bin_dir, 'schedule_logo.png')


def get_allocation_logo(bin_dir):
    return os.path.join(bin_dir, 'AllocationLogo.gif')



from tkinter import *
from src.scheduling_and_allocation.modified_tk import set_default_fonts_and_colours


# TODO: background button Colour changes don't work on mac, not yet tested on windows

def main():
    mw: Tk = Tk()
    colours = set_default_fonts_and_colours(mw, invert=True)
    mw.geometry('400x450')
    mw.title('Button Background Example')

    button = Button(mw, text='Submit', bg='blue', fg='pink', activebackground="red")
    button.pack()

    b1 = Button(mw, width=20, text='Button', bg='red')
    b2 = Text(mw, width=40, height=20)
    b2.insert('end', 'background colours of buttons have been set to blue and red')

    b1.pack()
    b2.pack()
    # b1.configure({'bg':"#000000"})

    print(f"Button background : {b1.cget('background')}")
    print(f"Button foreground : {b1.cget('foreground')}")

    mw.mainloop()


main()



"""Dialog box to edit a 'section' object"""
from __future__ import annotations

import tkinter as tk
import tkinter.font as tkFont
from tkinter.simpledialog import Dialog
from typing import Callable

from ..gui_generics.number_validations import validate_int, entry_int
from ..gui_dialogs.dialog_utilities import validate_class_times_equals_course_time, \
    get_block_info_from_tk_widgets, refresh_gui_blocks


# =====================================================================================================================
# Add a section to an existing course
# =====================================================================================================================
class AddSectionDialogTk(Dialog):
    def __init__(self, frame:tk.Frame,
                 *,
                 course_description: str,
                 course_hours: float,
                 apply_changes: Callable[[int, list], None]):
        """
        :param frame: parent frame of this dialog box
        :param course_description: The name of the course
        :param course_hours: How many hours is this course per week
        :param apply_changes: Who ya gonna call (Ghost Busters) to update the schedule with the users' changes
        """

        self.block_tk_variables = []
        self.course_hours = course_hours
        self.current_blocks =  []
        self.block_frames = None
        self.course_description = course_description
        self.top_frame = frame
        self._apply_changes = apply_changes

        self.description = tk.StringVar(value = "")
        self.number_of_sections = tk.StringVar(value="1")

        dialog_title = "Add Section(s)"
        super().__init__(frame.winfo_toplevel(), dialog_title)

    # ================================================================================================================
    # The content of the main body of the dialog box
    # ================================================================================================================
    def body(self, frame:tk.Frame):

        # ------------------------------------------------------------------------------------------------------------
        # Info
        # ------------------------------------------------------------------------------------------------------------
        course_info_frame = tk.Frame(frame)
        section_info_frame = tk.Frame(frame)
        lbl = tk.Label(course_info_frame,text=self.course_description, anchor='center',width=20)
        lbl.pack(expand=1,fill='both',padx=15,pady=5)
        default_font = tkFont.nametofont(lbl.cget("font"))
        family = default_font["family"]
        size = default_font["size"] + 2
        lbl.config(font=(family, size))

        tk.Label(section_info_frame, text="How many sections?", anchor='e', width=20).pack(side='left', padx=10, pady=5)
        en_number = entry_int(section_info_frame, textvariable=self.number_of_sections)
        en_number.pack(side='left', padx=10, pady=5)

        # ------------------------------------------------------------------------------------------------------------
        # Blocks ... there are no blocks, so just set up the frame
        # ------------------------------------------------------------------------------------------------------------
        self.block_frames = tk.Frame(frame)

        # ------------------------------------------------------------------------------------------------------------
        # layout
        # ------------------------------------------------------------------------------------------------------------
        course_info_frame.grid(row=0, column=0)
        section_info_frame.grid(row=1, column = 0)
        self.block_frames.grid(row=2, column=0, sticky='nsew', padx=5, pady=5)

        self.refresh()

        return en_number

    # ================================================================================================================
    # refresh blocks
    # ================================================================================================================
    def add_new_block(self):
        """
        Add a new block (class time) to the section
        """
        self.block_tk_variables.append((tk.StringVar(value="Monday"),tk.StringVar(value="8.0"), tk.StringVar(value="1.5")))
        self.refresh()

    def refresh(self):
        """
        redraw all the blocks (class times)
        """
        refresh_gui_blocks(self)
        tk.Button(self.block_frames, text="Add New Class Time", command=self.add_new_block,padx=5).pack(expand=1, fill='y')


    # ================================================================================================================
    # are blocks set up properly?
    # ================================================================================================================
    def validate(self):
        """
        Is the data, as entered by the user, valid?
        :return: True if data is good (changes are applied), false otherwise (nothing happens)
        """
        if not validate_int(self.number_of_sections.get(),"Number of Sections", "Number of sections must be a number!"):
            return False

        if not validate_class_times_equals_course_time(self.block_tk_variables, float(self.course_hours)):
            return False

        return True

    # ================================================================================================================
    # apply changes
    # ================================================================================================================
    def apply(self):
        """apply the changes and close the dialog"""
        new_blocks=get_block_info_from_tk_widgets(self.block_tk_variables)
        self._apply_changes( int(self.number_of_sections.get()), new_blocks )





"""
    Scheduler and Allocation - create teacher/lab/course schedules
    Copyright (C) 2025  Sandy Bultena

    This program comes with ABSOLUTELY NO WARRANTY.
    This is free software, and licensed under the GNU General Public License.
    (see <https://www.gnu.org/licenses/>)
"""
import os
import sys
import traceback

bin_dir: str = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(bin_dir, "../"))

def AllocationManager():
    try:
        from scheduling_and_allocation.presenter.allocation_manager import AllocationManager
        AllocationManager(bin_dir=bin_dir)
    except Exception as e:
        print()
        print("==================================================================")
        print("ERROR:")
        print("==================================================================")
        print(e)
        print()
        print("==================================================================")
        print("Trace")
        print("==================================================================")
        traceback.print_exc()
        print()
        print("==================================================================")
        print("System info")
        print("==================================================================")
        print(sys.version)
        print(sys.version_info)
        print()
        print()
        _=input("press enter to continue")


if __name__ == "__main__":
    AllocationManager()



from __future__ import annotations
from typing import Callable

import pytest

from src.scheduling_and_allocation.Utilities.id_generator import IdGenerator
from src.scheduling_and_allocation.model import Lab, Stream, Teacher, Block, ResourceType, ConflictType, Schedule, \
    SemesterType, WeekDay
from src.scheduling_and_allocation.presenter.views_controller import ViewsController

# =====================================================================================================================
# Dummy classes
# =====================================================================================================================
"""Provides code to deal with user modifying the gui view.  Most actions are passed onto the View Controller"""


RESOURCE = Lab | Stream | Teacher

# =====================================================================================================================
# use an generator to always guarantee that the new id will be unique
# =====================================================================================================================
_gui_block_ids = IdGenerator()

# =====================================================================================================================
# ViewTest
# =====================================================================================================================
class GUI:
    def raise_to_top(self): pass

VIEW_TEST_CALLED = 0
REFRESH_COLOURS_CALLED = 0
MOVE_GUI_BLOCK_TO = 0
VIEW_DRAW_CALLED = 0
class ViewTest:

    def __init__(self, views_controller, frame, schedule, resource, gui=None):
        self.gui = GUI()
        global VIEW_TEST_CALLED
        VIEW_TEST_CALLED += 1
        self.gui_blocks = {}
        self.move_gui_block_args = []

    def move_gui_block_to(self, block: Block, day: float, start_time: float):
        global MOVE_GUI_BLOCK_TO
        self.move_gui_block_args.append((block, day, start_time))
        MOVE_GUI_BLOCK_TO += 1

    def refresh_block_colours(self):
        global REFRESH_COLOURS_CALLED
        REFRESH_COLOURS_CALLED += 1

    def get_block_text(self, block: Block):
        return "hi"

    def draw(self):
        global VIEW_DRAW_CALLED
        VIEW_DRAW_CALLED += 1

    def is_block_in_view(self, block):
        return True

# =====================================================================================================================
# ViewsControllerTestTkTest
# =====================================================================================================================

class ViewsControllerTkTest:
    def __init__(self, parent, resources:dict[ResourceType,list], btn_callback: Callable):
        self.btn_colours = []

    def set_button_colour(self, button_id: str, resource_type, view_conflict: ConflictType = None):
        self.btn_colours.append((button_id, resource_type, view_conflict))

# =====================================================================================================================
# fixtures
# =====================================================================================================================
@pytest.fixture()
def gui():
    return ViewsControllerTkTest("",{},lambda x: None)

@pytest.fixture()
def schedule_obj():
    schedule = Schedule()

    # teachers
    t1 = schedule.add_update_teacher("Jane", "Doe", "0.25", teacher_id="1")
    t2 = schedule.add_update_teacher("John", "Doe", teacher_id="2")
    t3 = schedule.add_update_teacher("Babe", "Ruth", teacher_id="3")
    t4 = schedule.add_update_teacher("Bugs", "Bunny", teacher_id = "4")

    # labs
    l1 = schedule.add_update_lab("P107", "C-Lab")
    l2 = schedule.add_update_lab("P322", "Mac Lab")
    l3 = schedule.add_update_lab("P325")
    l4 = schedule.add_update_lab("BH311", "Britain Hall")

    # streams
    st1 = schedule.add_update_stream("1A", "calculus available")
    st2 = schedule.add_update_stream("1B", "no calculus")
    st3 = schedule.add_update_stream("2A")
    st4 = schedule.add_update_stream("2B")

    # courses
    c_001 = schedule.add_update_course("001","BasketWeaving", SemesterType.fall )
    s_001_1 = c_001.add_section("1", section_id=1)
    s_001_1.add_stream(st1)
    b_001_1_1 = s_001_1.add_block(WeekDay.Monday,  8 )
    b_001_1_2 = s_001_1.add_block(WeekDay.Monday,  10 )
    b_001_1_1.add_teacher(t1)
    b_001_1_2.add_teacher(t1)
    b_001_1_1.add_lab(l1)
    b_001_1_2.add_lab(l1)

    s_001_2 = c_001.add_section("2",section_id=2)
    s_001_2.add_stream(st2)
    b_001_2_1 = s_001_2.add_block(WeekDay.Tuesday,  8 )
    b_001_2_2 = s_001_2.add_block(WeekDay.Tuesday,  10 )
    b_001_2_1.add_teacher(t2)
    b_001_2_2.add_teacher(t3)
    b_001_2_1.add_lab(l2)
    b_001_2_2.add_lab(l3)


    c_002 = schedule.add_update_course("002","Thumb Twiddling", SemesterType.fall )
    s_002_1 = c_002.add_section("1",)
    b_002_1_1 = s_002_1.add_block(WeekDay.Monday,  8 )
    b_002_1_2 = s_002_1.add_block(WeekDay.Monday,  10 )
    s_002_2 = c_002.add_section("2",)
    b_002_2_1 = s_002_2.add_block(WeekDay.Tuesday,  8 )
    b_002_2_2 = s_002_2.add_block(WeekDay.Tuesday,  10 )



    return schedule

DIRTY_FLAG = False
def dirty_flag_method(value=None):
    global DIRTY_FLAG
    if value is not None:
        DIRTY_FLAG = value
    return DIRTY_FLAG

# =====================================================================================================================
# Testing
# =====================================================================================================================
def test_init(schedule_obj, gui):
    """create views controller
    1. all resources are saved in object
    """
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)
    assert vc.resources[ResourceType.teacher] == list(schedule_obj.teachers())
    assert vc.resources[ResourceType.lab] == list(schedule_obj.labs())
    assert vc.resources[ResourceType.stream] == list(schedule_obj.streams())

def test_create_view_only_once(schedule_obj, gui, monkeypatch):
    """create multiple views, but only end up with single view for each resource"""

    # prepare
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)
    teacher1: Teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    teacher2: Teacher = schedule_obj.get_teacher_by_name("John","Doe")

    # execute
    vc.call_view(teacher1)
    vc.call_view(teacher2)
    vc.call_view(teacher1)
    vc.call_view(teacher2)
    vc.call_view(teacher1)
    vc.call_view(teacher2)

    # validate
    assert len(vc._views) == 2


def test_refresh(schedule_obj, gui):
    """refresh views controller
    1. conflicts are recalculated
    2. all buttons in resources have their colour updated
    """

    # prepare
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)
    teacher: Teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    blocks = schedule_obj.get_blocks_for_teacher(teacher)
    block1 = blocks[0]
    block2 = blocks[1]
    block1.day = block2.day
    block1.start = block2.start
    assert not block1.conflict
    assert not block2.conflict

    # execute
    vc.refresh()

    # validate
    assert block1.conflict
    assert block2.conflict
    assert (len(gui.btn_colours) == len(vc.resources[ResourceType.teacher])
            + len(vc.resources[ResourceType.lab])
            + len(vc.resources[ResourceType.stream])
            )


def test_block_moved(schedule_obj, gui, monkeypatch):
    global REFRESH_COLOURS_CALLED
    """block has moved
    1. dirty flag has not been reset (it is reset when the block is dropped)
    2. refresh was called (which recalculates conflicts)
    3. all currently open views colours are updated 
    4. views are told to update their gui_block positions (not tested)
    """
    # prepare
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)
    teacher: Teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    teacher1: Teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    teacher2: Teacher = schedule_obj.get_teacher_by_name("John","Doe")
    blocks = schedule_obj.get_blocks_for_teacher(teacher)
    block1 = blocks[0]
    block2 = blocks[1]
    dirty_flag_method(False)
    vc.call_view(teacher1)
    vc.call_view(teacher2)
    assert not block1.conflict
    assert not block2.conflict

    # execute
    REFRESH_COLOURS_CALLED = 0
    vc.notify_block_move("", block1, block1.day.value, block1.start)

    # verify
    assert not dirty_flag_method()
    assert block1.conflict
    assert block2.conflict
    assert REFRESH_COLOURS_CALLED == 2

def test_block_moved_view_updated(schedule_obj, gui, monkeypatch):
    global MOVE_GUI_BLOCK_TO
    """block has moved
    1. views are told to update their gui_block positions (but not the one that triggered it)
    """

    # prepare
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)

    # both teacher and stream share the same blocks
    teacher: Teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    stream: Stream = schedule_obj.get_stream_by_number("1A")
    blocks = schedule_obj.get_blocks_for_teacher(teacher)
    block1, block2 = blocks
    vc.call_view(teacher)
    vc.call_view(stream)

    # execute
    MOVE_GUI_BLOCK_TO = 0
    vc.notify_block_move(teacher.number, block1, block1.day.value, block1.start)

    # verify
    assert MOVE_GUI_BLOCK_TO == 1

def test_block_movable_changed(schedule_obj, gui, monkeypatch):
    """block movable has been changed
    1. NOTE: dirty flag is updated in the view where the event happened
    2. update colours on all the views
    3. update colors on view choices
    4. conflicts are recalculated
    """
    global REFRESH_COLOURS_CALLED
    REFRESH_COLOURS_CALLED = 0

    # prepare
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)
    teacher: Teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    stream: Stream = schedule_obj.get_stream_by_number("1A")
    blocks = schedule_obj.get_blocks_for_teacher(teacher)
    block1, block2 = blocks
    vc.call_view(teacher)
    vc.call_view(stream)
    assert not block1.conflict

    # execute
    block1.movable = not block1.movable
    vc.notify_block_movable_toggled(block1)

    # validate
    assert block1.conflict

    # ... validate update colours on all the open views
    assert REFRESH_COLOURS_CALLED == 2

    # ... validate update colours on all the views
    assert (len(gui.btn_colours) == len(vc.resources[ResourceType.teacher])
            + len(vc.resources[ResourceType.lab])
            + len(vc.resources[ResourceType.stream])
            )


def test_block_moved_to_different_resource1(schedule_obj, gui, monkeypatch):
    """block is to be moved to a different resource
    1. block is removed from original resource
    2. block is added to the other resource
    3. dirty flag is set
    4. conflicts are updated
    5. conflicts are recalculated
    """
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)

    # prepare
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)
    teacher: Teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    teacher2: Teacher = schedule_obj.get_teacher_by_name("John", "Doe")
    blocks = schedule_obj.get_blocks_for_teacher(teacher)
    block1, block2 = blocks

    # execute
    dirty_flag_method(False)
    block2.conflict = ConflictType.NONE
    vc.notify_move_block_to_resource(ResourceType.teacher, block1, teacher, teacher2)

    # validate
    assert block1 not in schedule_obj.get_blocks_for_teacher(teacher)
    assert block1 in schedule_obj.get_blocks_for_teacher(teacher2)
    assert dirty_flag_method()
    assert block2.conflict

def test_block_moved_to_different_resource2(schedule_obj, gui, monkeypatch):
    """block is to be moved to a different resource, both views are already opened
    1. both views are redrawn
    """
    global VIEW_DRAW_CALLED
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)

    # prepare
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)
    teacher: Teacher = schedule_obj.get_teacher_by_name("Jane","Doe")
    teacher2: Teacher = schedule_obj.get_teacher_by_name("John", "Doe")
    blocks = schedule_obj.get_blocks_for_teacher(teacher)
    block1, block2 = blocks
    vc.call_view(teacher)
    vc.call_view(teacher2)
    VIEW_DRAW_CALLED = 0

    # execute
    vc.notify_move_block_to_resource(ResourceType.teacher, block1, teacher, teacher2)

    # validate
    assert VIEW_DRAW_CALLED == 2


def test_block_moved_to_different_resource3(schedule_obj, gui, monkeypatch):
    """block is to be moved to a different resource, only one view is already open
    1. one view is redrawn, the other is created
    """
    global VIEW_DRAW_CALLED
    global VIEW_TEST_CALLED
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)

    # prepare
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)
    teacher: Teacher = schedule_obj.get_teacher_by_name("Jane", "Doe")
    teacher2: Teacher = schedule_obj.get_teacher_by_name("John", "Doe")
    blocks = schedule_obj.get_blocks_for_teacher(teacher)
    block1, block2 = blocks
    vc.call_view(teacher)
    VIEW_DRAW_CALLED = 0
    VIEW_TEST_CALLED = 0

    # execute
    vc.notify_move_block_to_resource(ResourceType.teacher, block1, teacher, teacher2)

    # validate
    assert VIEW_DRAW_CALLED == 1
    assert VIEW_TEST_CALLED == 1

def test_undos_from_saved_actions(schedule_obj, gui, monkeypatch):
    """create 3 actions of type move, change resource, change movability
    1. undo list should have three actions saved
    2. redo list should have no actions saved
    """
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)
    teacher: Teacher = schedule_obj.get_teacher_by_name("Jane", "Doe")
    teacher2: Teacher = schedule_obj.get_teacher_by_name("John", "Doe")
    blocks = schedule_obj.get_blocks_for_teacher(teacher)
    block1, block2 = blocks

    # execute
    vc.save_action_block_move(block1, from_day=5, to_day=block1.day.value,
                              from_time=0, to_time=block1.start)
    vc.save_action_block_resource_changed(ResourceType.teacher, block1, from_resource=teacher2, to_resource=teacher)
    vc.save_action_block_movable_toggled(block2,  block2.movable)

    # validate
    undos = vc._undo
    assert len(undos) == 3
    assert len(vc._redo) == 0
    a1,a2,a3 = undos
    assert a1.action == "move"
    assert a1.from_day == 5
    assert a1.to_day == block1.day.value
    assert a1.from_time == 0
    assert a1.to_time == block1.start

    assert a2.action == "change_resource"
    assert a2.to_resource == teacher
    assert a2.from_resource == teacher2

    assert a3.action == "toggle_movable"
    assert a3.was_movable == block2.movable

def test_undos_actual_undo(schedule_obj, gui, monkeypatch):
    """create 3 actions, then undo all changes, one by one
    1. undo action is done
    2. undo action is saved in redo
    """
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)
    teacher: Teacher = schedule_obj.get_teacher_by_name("Jane", "Doe")
    teacher2: Teacher = schedule_obj.get_teacher_by_name("John", "Doe")
    blocks = schedule_obj.get_blocks_for_teacher(teacher)
    block1, block2 = blocks

    original_movable = block2.movable
    block2.movable = not block2.movable

    vc.save_action_block_move(block1, from_day=5, to_day=block1.day.value,
                              from_time=14, to_time=block1.start)
    vc.save_action_block_resource_changed(ResourceType.teacher, block1, from_resource=teacher2, to_resource=teacher)
    vc.save_action_block_movable_toggled(block2,  original_movable)

    # execute
    vc.undo()

    # validate
    assert len(vc._undo) == 2
    assert len(vc._redo) == 1
    assert block2.movable == original_movable

    # execute
    vc.undo()

    # validate
    assert len(vc._undo) == 1
    assert len(vc._redo) == 2
    assert block1 in schedule_obj.get_blocks_for_teacher(teacher2)

    # execute
    vc.undo()

    # validate
    assert len(vc._undo) == 0
    assert len(vc._redo) == 3
    assert block1.day.value == 5
    assert block1.start == 14

def test_redos(schedule_obj, gui, monkeypatch):
    """create 3 actions, then undo all changes, and then redo them
    1. after, everything is back to the original
    2. redo action is saved in 'undo'
    """
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)
    teacher: Teacher = schedule_obj.get_teacher_by_name("Jane", "Doe")
    teacher2: Teacher = schedule_obj.get_teacher_by_name("John", "Doe")
    blocks = schedule_obj.get_blocks_for_teacher(teacher)
    block1, block2 = blocks

    original_time = block1.start
    original_movable = block2.movable
    block2.movable = not block2.movable

    vc.save_action_block_move(block1, from_day=5, to_day=block1.day.value,
                              from_time=14, to_time=block1.start)
    vc.save_action_block_resource_changed(ResourceType.teacher, block1, from_resource=teacher2, to_resource=teacher)
    vc.save_action_block_movable_toggled(block2,  original_movable)

    # execute
    vc.undo()
    vc.undo()
    vc.undo()
    vc.redo()
    vc.redo()
    vc.redo()

    # validate
    assert len(vc._undo) == 3
    assert len(vc._redo) == 0
    assert block2.movable != original_movable
    assert block1 in schedule_obj.get_blocks_for_teacher(teacher)
    assert block1.start == original_time

def test_undo_redo_no_action_required(schedule_obj, gui, monkeypatch):
    """no pending actions,
    1. redo/undo do not crash the program
    """
    monkeypatch.setattr("src.scheduling_and_allocation.presenter.views_controller.View", ViewTest)
    vc = ViewsController(dirty_flag_method, "", schedule_obj, gui)

    # execute
    vc.undo()
    vc.undo()
    vc.undo()
    vc.redo()
    vc.redo()
    vc.redo()




import tkinter as tk
from tkinter import ttk, messagebox

from src.scheduling_and_allocation.modified_tk import DragNDropManager


def lb_on_start(e: tk.Event, info_data: dict) -> str:
    # select the nearest item in the list
    w: tk.Listbox = e.widget
    index = w.nearest(e.y)
    w.select_clear(0, 'end')
    w.selection_set(index, index)

    # save some info for later use
    info_data["index"] = index

    # return string to be used in the drag indicator
    return w.get(index)


def lb_on_drag(e: tk.Event, info_data: dict, target: tk.Widget) -> bool:
    for s_iid in tv.selection():
        tv.selection_remove(s_iid)

    # if we are not dropping on the treeview, then it is no good
    if target != tv:
        return False

    # made up rule, just to demonstrate use of info_data
    if info_data["index"] < 0:
        return False

    # select item are we hovering over?
    tv_y_pos = e.y_root - tv.winfo_rooty()
    iid = tv.identify_row(tv_y_pos)
    if not iid:
        return False

    # if selected item doesn't have a parent, then it is not a valid drop site, but open branch
    if tv.parent(iid) == "":
        tv.item(iid, open=True)
        return False

    # finally, we have a valid drop site
    tv.selection_set(iid)
    tv.item(iid, open=True)
    return True


def lb_on_drop(e: tk.Event, info_data: dict, target: tk.Widget):
    tv.selection_clear()

    # if we are not dropping on the treeview, then it is no good
    if target != tv:
        return

    # select item are we hovering over?
    tv_y_pos = e.y_root - tv.winfo_rooty()
    iid = tv.identify_row(tv_y_pos)
    if not iid:
        return

    # if selected item doesn't have a parent, then it is not a valid drop site, but open branch
    if tv.parent(iid) == "":
        return

    # finally, we have a valid drop site
    messagebox.showinfo(title="Successful Drop", message=f"index {info_data['index']} dropped on {tv.item(iid)}")


mw = tk.Tk()
mw.geometry("300x500+30+40")
#colours, fonts = set_default_fonts_and_colours(mw)
tk.Label(mw, text="Drag from list one onto tree").pack()

lb = tk.Listbox()
lb.pack()
lb.insert('end', "one")
lb.insert('end', "two")
lb.insert('end', "three")
lb.insert('end', "four")

tv = ttk.Treeview()
tv.pack()
id1 = tv.insert(parent="", index="end", text="Not droppable", open=False)
id2 = tv.insert(parent=id1, index="end", text="can drop on me location 1", open=False)
id3 = tv.insert(parent=id2, index="end", text="can drop on me location 2", open=False)

tk.Button(mw, text="clear selection", command=lambda: tv.selection_clear()).pack()

dm = DragNDropManager()
dm.add_draggable(lb, on_start=lb_on_start, on_drop=lb_on_drop, on_drag=lb_on_drag)
mw.mainloop()



"""
# ============================================================================
# Code that draws the View stuff.  No other functionality.
#
# PURPOSE: To provide a simple class that can be used to draw schedules where the canvas object can
#          be a generic canvas:  PDF, Tk.Canvas, HTML, etc.
#
# EVENT HANDLERS: None
# ============================================================================
"""

import tkinter as tk
from typing import Protocol, Optional

from ..Utilities import Colour
from ..gui_generics.block_colours import RESOURCE_COLOURS
from ..model import ResourceType
from ..gui_generics.drawing_scale import DrawingScale
from ..modified_tk.InitGuiFontsAndColours import get_fonts_and_colours

DAYS = ("Monday", "Tuesday", "Wednesday", "Thursday", "Friday")
Times: dict[int, str] = {
    8: "8am",
    9: "9am",
    10: "10am",
    11: "11am",
    12: "12pm",
    13: "1pm",
    14: "2pm",
    15: "3pm",
    16: "4pm",
    17: "5pm",
    18: "6pm"
}
RECTANGLE_X1_OFFSET = 3
RECTANGLE_Y1_OFFSET = 2
RECTANGLE_X2_OFFSET = -2
RECTANGLE_Y2_OFFSET = -3

EARLIEST_TIME = min(Times.keys())
LATEST_TIME = max(Times.keys())

# =====================================================================================================================
# what is the minimal requirements for the canvas object to have if we want to draw
# =====================================================================================================================
class GenericCanvas(Protocol):
    def __init__(self, title="Title", schedule_name="sub_title", filename=None):...
    def create_line(self, x1:float, y1:float, x2:float, y2:float, fill:str="grey", dash="", tags: str|tuple=""):...
    def addtag_withtag(self,*args,**kwargs):...
    def create_text(self, x:float, y:float, text:str="", fill:str='black', tags:str|tuple="")->int:...
    def create_rectangle(self, coords:tuple[float,float,float,float], fill:str='grey', outline:str='grey',
                                        tags:str|tuple="")->int:...
    def save(self):...


# =====================================================================================================================
# ViewCanvasTk: this is where we draw stuff :)
# =====================================================================================================================
class ViewCanvasTk:

    Movable_Tag_Name = "movable"
    Rectange_Tag_Name = "rectangle"
    Text_Tag_Name = "text"
    Clickable_Tag_Name  = "clickable"

    # =================================================================
    # __init__
    # =================================================================
    def __init__(self, canvas: tk.Canvas | GenericCanvas, scale_factor: float = 1,
                 fg_colour=None, bg_colour=None):
        """Draws the Schedule timetable on the specified canvas.
            :param canvas: what to draw on
            :param scale_factor: scaling factors
            :param bg_colour: the colour used to draw the objects
        """
        self.colours, self.fonts = get_fonts_and_colours()

        self.canvas = canvas
        self.scale = DrawingScale(scale_factor)
        self.scale_factor = DrawingScale(scale_factor)
        scale = self.scale
        self.bg_colour = self.colours.DataBackground if bg_colour is None else bg_colour
        self.fg_colour = self.colours.WindowForeground if fg_colour is None else fg_colour
        self.canvas.config(background=self.bg_colour)

        # --------------------------------------------------------------------
        # draw hourly lines
        # --------------------------------------------------------------------
        (dummy_x, x_max) = self._days_x_coords(len(DAYS))
        (x_min, dummy_y) = self._days_x_coords(1)

        for time in Times.keys():

            # Draw each hour line.
            (y_hour, y_half) = self._time_y_coords(time, 0.5)
            canvas.create_line(
                x_min, y_hour, x_max, y_hour,
                fill="dark grey",
                dash="-",
                tags="baseline",
            )

            # Hour text.
            canvas.create_text(
                (x_min + scale.x_origin) / 2,
                y_hour, text=Times[time],
                fill=self.fg_colour,
                tags="baseline",
            )

            # for all inner times, draw a dotted line for the half hour.
            if time != LATEST_TIME:
                canvas.create_line(
                    x_min, y_half, x_max, y_half,
                    fill="grey",
                    dash=".",
                    tags="baseline",
                )

                # Half-hour text.
                if scale.current_scale > 0.5:
                    canvas.create_text(
                        (x_min + scale.x_origin) / 2,
                        y_half, text=":30",
                        fill=self.fg_colour,
                        tags="baseline",
                    )

        # --------------------------------------------------------------------
        # draw day lines
        # --------------------------------------------------------------------
        (y_min, y_max) = self._time_y_coords(EARLIEST_TIME, (LATEST_TIME - EARLIEST_TIME))

        for i in range(len(DAYS) + 1):
            (x_day, x_day_end) = self._days_x_coords(i + 1)
            canvas.create_line(x_day, scale.y_height_scale, x_day, y_max, fill=self.fg_colour,tags="baseline",)

            # day text
            if i < len(DAYS):
                if scale.current_scale <= 0.5:
                    canvas.create_text(
                        (x_day + x_day_end) / 2,
                        (y_min + scale.y_origin) / 2,
                        text=DAYS[i][0:1],
                        fill=self.fg_colour,
                        tags=("baseline",),
                    )
                else:
                    canvas.create_text(
                        (x_day + x_day_end) / 2,
                        (y_min + scale.y_origin) / 2,
                        text=DAYS[i],
                        fill=self.fg_colour,
                        tags=("baseline",),
                    )

    # =================================================================
    # draw_block
    # =================================================================
    def draw_block(self, resource_type: ResourceType, day: int, start_time: float,
                   duration: float, text:str, gui_tag, movable):
        """
        Draws the specific block on the gui canvas
        :param resource_type:
        :param day:
        :param start_time:
        :param duration:
        :param text:  the text to display on the block
        :param gui_tag:  the unique identifier for all gui objects that comprise this block
        :return:
        """

        # colour
        colour = RESOURCE_COLOURS[resource_type]
        colour = Colour.string(colour)
        text_colour = "black"
        if not Colour.is_light(colour):
            text_colour = "white"

        # coordinates
        x1,y1,x2,y2 = self.get_coords(day, start_time, duration)

        # draw
        self.canvas.create_rectangle(
            (x1 + RECTANGLE_X1_OFFSET,
             y1 + RECTANGLE_Y1_OFFSET,
             x2 + RECTANGLE_X2_OFFSET,
             y2 + RECTANGLE_Y2_OFFSET),
            fill=colour, outline=colour,
            tags=(self.Rectange_Tag_Name, self.Clickable_Tag_Name, gui_tag))

        self.canvas.create_text(
            (x1 + x2) / 2, (y1 + y2) / 2, text=text, fill=text_colour,
            tags=(self.Text_Tag_Name, self.Clickable_Tag_Name, gui_tag)
        )
        #self.canvas.addtag_withtag(self.Movable_Tag_Name, gui_tag)
        if movable:
            self.canvas.addtag_withtag(self.Movable_Tag_Name, gui_tag)
        else:
            self.canvas.tag_raise(gui_tag,"baseline")

    # =================================================================
    # modify movability of gui block
    # =================================================================
    def modify_movable(self, gui_tag:str, movable: bool):
        """
        modify movability of gui block
        :param gui_tag:
        :param movable:
        :return:
        """
        if movable:
            self.canvas.addtag_withtag(self.Movable_Tag_Name, gui_tag)
        else:
            self.canvas.dtag(gui_tag, self.Movable_Tag_Name)

    # =================================================================
    # get the gui id from tags
    # =================================================================
    def get_gui_block_id_from_selected_item(self) -> Optional[str]:
        """
        From the currently selected item, return the gui block tag
        :return: None if none can be found, else remaining tags, anded together
        """
        selected_objects: tuple[int, ...] = self.canvas.find_withtag(f"current && {self.Clickable_Tag_Name}")
        if len(selected_objects) == 0:
            return None

        selected_object = selected_objects[0]

        all_tags = set(self.canvas.gettags(selected_object)) - {'current', self.Rectange_Tag_Name,
                                                            self.Text_Tag_Name,
                                                            self.Movable_Tag_Name,
                                                            self.Clickable_Tag_Name}
        if len(all_tags) == 0:
            return ""
        if len(all_tags) > 1:
            print("SOMETHING IS WRONG IN VIEW_CANVAS_TK", all_tags)
        gui_id = list(all_tags)[0]
        return gui_id

    # =================================================================
    # get coordinates for a gui block
    # =================================================================
    def gui_block_coords(self, gui_block_id) -> Optional[list[float]]:
        """
        returns the coordinates of the gui block, None if gui block doesn't exist
        :param gui_block_id:
        :return: a list of 4 floats, or 'None'
        """
        obj_id = self.canvas.find_withtag(f"{ViewCanvasTk.Rectange_Tag_Name} && {gui_block_id}")
        if len(obj_id) < 1:
            return None
        else:
            coords = self.canvas.coords(obj_id[0])
            return [coords[0]-RECTANGLE_X1_OFFSET, coords[1]-RECTANGLE_Y1_OFFSET,
                    coords[2]-RECTANGLE_X2_OFFSET, coords[3]-RECTANGLE_Y2_OFFSET]

    # =================================================================
    # coords_to_day_time_duration
    # =================================================================
    def coords_to_day_time_duration(self, x, y, y2) -> tuple[float, float, float]:
        """
        Determines the day, time_start time, and duration based on canvas coordinates.
        :param x: x position (determines day)
        :param y: y position (determines time_start)
        :param y2: y2 position (with y - determines duration)
        :return: day, time, duration
        """
        scale = self.scale
        day = x / scale.x_width_scale - scale.x_offset + 1 - scale.x_origin
        time = y / scale.y_height_scale - scale.y_offset + EARLIEST_TIME - scale.y_origin
        duration = (y2 + 1 - y) / scale.y_height_scale

        return day, time, duration

    # =================================================================
    # gui block to day, time, duration (float)
    # =================================================================
    def gui_block_to_day_time_duration(self, gui_block_id) ->Optional[tuple[float,float,float]]:
        """
        Gets day/time/duration for given gui block if block exists
        :param gui_block_id:
        :return: day, start_time, duration if gui block exists, else None
        """

        coords = self.gui_block_coords(gui_block_id)
        if coords is not None:
            x1, y1, _, y2 = coords
            return self.coords_to_day_time_duration(x1, y1, y2)
        return None

    # =================================================================
    # get_coords
    # =================================================================
    def get_coords(self, day:float, start: float, duration: float=1.0):
        """Determines the canvas coordinates based on day, time_start time, and duration.
            :param start:
            :param day:
            :param duration:
        """
        (x, x2) = self._days_x_coords(round(day))
        (y, y2) = self._time_y_coords(start, duration)

        return x, y, x2, y2

    # =================================================================
    # using scale info, get the y limits for a specific time period
    # =================================================================
    def _time_y_coords(self, start, duration):
        """using scale info, get the y limits for a specific time period."""

        y_offset = self.scale.y_offset * self.scale.y_height_scale + self.scale.y_origin
        y = y_offset + (start - EARLIEST_TIME) * self.scale.y_height_scale
        y2 = duration * self.scale.y_height_scale + y - 1

        return y, y2


    # =================================================================
    # using scale info, get the x limits for a specific day
    # =================================================================
    def _days_x_coords(self, day: int) -> tuple[float,float]:
        """
        what can be the maximum and minimum number for day
        :param day:
        :return: minimum_x, maximum_x
        """

        x_offset = self.scale.x_offset * self.scale.x_width_scale + self.scale.x_origin
        x = x_offset + (day - 1) * self.scale.x_width_scale
        x2 = x_offset + day * self.scale.x_width_scale - 1

        return x, x2


    # ==================================================================
    # update scale
    # ==================================================================
    def adjust_scale(self, factor):
        """change the scale information based on the scaling factor"""
        self.scale = DrawingScale(factor)




from tkinter import *
from tkinter import messagebox
from typing import Any

from src.scheduling_and_allocation.gui_generics.menu_and_toolbars import MenuItem, MenuType
from src.scheduling_and_allocation.gui_pages import EditCoursesTk
from src.scheduling_and_allocation.model import ResourceType


# ----------------------------------------------------------------------------
# generic objects for putting in listbox and trees
# ----------------------------------------------------------------------------
class Nothing:
    def __init__(self, a):
        self.a = a

    def checking(self):
        print(f"object: a = {self.a}")

    def __str__(self):
        return str(self.a)

    def __repr__(self):
        return str(self.a)

    def __lt__(self, other):
        return self.a < other.a


# ----------------------------------------------------------------------------
# methods for use in callback methods
# ----------------------------------------------------------------------------
def tree_menu(obj) -> list[MenuItem]:
    menu = MenuItem(name=str(obj),
                    label=f"Testing: {str(obj)}",
                    menu_type=MenuType.Command,
                    command=lambda: None)
    return [menu, ]


def resource_menu(resource_type: ResourceType, obj) -> list[MenuItem]:
    menu_title = MenuItem(name=str(resource_type),
                          label=f"Testing: {str(resource_type)}",
                          menu_type=MenuType.Command,
                          command=lambda: None)
    menu = MenuItem(name=str(obj),
                    label=f"Testing: {str(obj)}",
                    menu_type=MenuType.Command,
                    command=lambda: None)
    return [menu_title, menu]


def edit_dialog(obj: Any, parent_obj: Any, tree_id: str, parent_id: str):
    messagebox.showinfo(title="Testing", message=f"selected obj: {obj}\n"
                                                 f"parent obj: {parent_obj}\n"
                                                 f"tree id: {tree_id}\n"
                                                 f"parent id: {parent_id}")


def new_course_dialog():
    messagebox.showinfo(title="Testing", message="New Course dialog")


def teacher_stat_dialog(obj: Any):
    messagebox.showinfo(title="Testing", message=f"Teacher stat for {str(obj)}")


def valid_drop_site(resource_type: ResourceType, target: Any) -> bool:
    if str(target) == "Sandy":
        return False
    if str(target).startswith("P"):
        return False
    return True


def item_dropped(source: Any, target: Any, id):
    messagebox.showinfo(title="Testing",
                        message=f"'{str(source)}' dropped on '{str(target)}' tree id '{str(id)}'")


# ----------------------------------------------------------------------------
# setup for creating the EditCourseTk page
# ----------------------------------------------------------------------------
mw = Tk()
f = Frame(mw)
f.pack(expand=1, fill="both")

test_page = EditCoursesTk(f)
t1 = Nothing("Sandy")
t2 = Nothing("Bob")
t3 = Nothing("Sandy1")
t4 = Nothing("Bob1")
t5 = Nothing("Sandy2")
t6 = Nothing("Bob2")
t7 = Nothing("Sandy3")
t8 = Nothing("Bob3")
t9 = Nothing("Sandy4")
t10 = Nothing("Bob4")
t11 = Nothing("Sandy5")
t12 = Nothing("Bob5")
l1 = Nothing("P325")
l2 = Nothing("P322")
l3 = Nothing("P326")
l4 = Nothing("P323")
s1 = Nothing("1A")
s2 = Nothing("1B")
s3 = Nothing("2A")
s4 = Nothing("2B")
s5 = Nothing("3A")
s6 = Nothing("3B")

test_page.update_resource_type_objects(ResourceType.teacher, [t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12])
test_page.update_resource_type_objects(ResourceType.stream, [s1, s2, s3, s4, s5, s6])
test_page.update_resource_type_objects(ResourceType.lab, [l1, l2, l3, l4])


c1 = Nothing("Operating Systems")
c1_id = test_page.add_tree_item("", str(c1), c1)
s1_id = test_page.add_tree_item(c1_id, f"Section 1 ({s1})", Nothing("Section 1"))
s2_id = test_page.add_tree_item(c1_id, f"Section 2 ({s2})", Nothing("Section 2"))
b1_id = test_page.add_tree_item(s1_id, "Block 1", Nothing("Block 1"))
b2_id = test_page.add_tree_item(s1_id, "Block 2", Nothing("Block 2"))
test_page.add_tree_item(b1_id, "Teacher: " + str(t1), t1)
test_page.add_tree_item(b1_id, "Lab: " + str(l1), l1)

# ----------------------------------------------------------------------------
# add all the callback routines
# ----------------------------------------------------------------------------
# call back routines
test_page.handler_tree_edit = edit_dialog
test_page.handler_new_course = new_course_dialog
test_page.handler_tree_create_popup = tree_menu
test_page.handler_resource_create_menu = resource_menu
test_page.handler_show_teacher_stat = teacher_stat_dialog
test_page.handler_drag_resource = valid_drop_site
test_page.handler_drop_resource = item_dropped

mw.mainloop()



"""
# ============================================================================
# Basically a Tk::TableEntry object with some restrictions
#
# The first column I<must> be a unique identifier for the corresponding data
# object, and can not be edited.
#
# EVENT HANDLERS
#
#   button-click: delete the data in this row (code is in Tk::TableEntry)
#       event_delete_handler(row_data)
#
#   leave form event: save the data (data is a 2d list of strings)
#       event_save_handler(all_data)
#
# ============================================================================

"""
from asyncio import sleep
from typing import *

from ..Utilities import Preferences
from ..modified_tk.TableEntry import TableEntry
from dataclasses import dataclass
import tkinter as tk
from ..modified_tk.InitGuiFontsAndColours import TkColours


# ============================================================================
# Class for describing properties for each column
# ============================================================================
@dataclass
class DEColumnDescription:
    title: str
    width: int
    property: str
    unique_id: bool


# ============================================================================
# class: EditResourcesTk
# ============================================================================
class EditResourcesTk:
    Currently_saving = 0

    # ------------------------------------------------------------------------
    # constructor
    # ------------------------------------------------------------------------
    def __init__(self,
                 parent: tk.Frame,
                 event_delete_handler: Callable[[list[str], ...], None] = lambda x, *_: None,
                 event_save_handler: Callable[[list[list[str]]], None] = lambda *_: None,
                 preferences: Optional[Preferences] = None,
                 colours: Optional[TkColours] = None,
                 ):
        """
        :param parent: the frame where to put the data
        :param event_delete_handler: function to call when something is deleted
        :param event_save_handler: function to call for saving data
        :param colours: colours
        """
        # remove anything that was already in the frame
        for widget in parent.winfo_children():
            widget.destroy()

        if colours is None:
            colours = TkColours(parent.winfo_toplevel(), preferences.dark_mode())
        self.colours = colours
        self.frame = parent
        self.data_entry: Optional[TableEntry] = None
        self.delete_handler = event_delete_handler
        self.save_handler = event_save_handler

    # ------------------------------------------------------------------------
    # initialize columns
    # ------------------------------------------------------------------------
    def initialize_columns(self, column_descriptions: list[DEColumnDescription], disabled_columns=None):
        """
        define the columns
        :param column_descriptions:
        :param disabled_columns:
        :return:
        """

        titles: list[str] = list()
        column_widths: list[int] = list()
        for cd in column_descriptions:
            titles.append(cd.title)
            column_widths.append(cd.width)

        if self.data_entry is not None:
            self.data_entry.destroy()

        self.data_entry = TableEntry(
            self.frame,
            rows=1,
            columns=len(column_descriptions),
            titles=titles,
            colwidths=column_widths,
            delete=self.delete_handler,
            colours=self.colours,
            disabled = disabled_columns
        )
        self.data_entry.pack(side='top', expand=True, fill='both')

        self.data_entry.bind('<Leave>', func=self.save)

    # ------------------------------------------------------------------------
    # refresh
    # ------------------------------------------------------------------------
    def refresh(self, data: list[list[Any]]):
        """
        refresh the data
        :param data: a list of lists (a 2-d array sort of)
        """
        self.data_entry.clear_data()
        for row, row_data in enumerate(data):
            for col, cell_data in enumerate(row_data):
                self.data_entry.put(row, col, str(cell_data))

        if self.data_entry.number_of_rows <= len(data):
            self.data_entry.add_empty_row()

        # stupid Tk won't update image unless I do this.  Oh well, at least it worked
        self.frame.focus_set()

    # ------------------------------------------------------------------------
    # get the data from the gui and return it
    # ------------------------------------------------------------------------
    def get_all_data(self) -> list[list[str]]:
        """
        reads the data stored in the TableEntry widget
        :return: a list of rows, which is a list of columns with the data
        """
        data: list = list()
        for row in range(self.data_entry.number_of_rows):
            data.append(self.data_entry.read_row(row))
        return data

    # ------------------------------------------------------------------------
    # save
    # ------------------------------------------------------------------------
    def save(self, *_):
        """save the data in the table"""
        # Just in case saving is already in progress, wait before continuing.
        if self.Currently_saving > 2:
            return  # 2 is too many.
        if self.Currently_saving:
            sleep(2)
        self.Currently_saving += 1

        # call event handler with data
        all_data = self.get_all_data()
        self.save_handler(all_data)

        # reset currently saving
        self.Currently_saving -= 1




"""
# ============================================================================
# This is the central control for launching any or all of the Views
# (the things that allow you to move blocks around)
#
# For each schedule item view, there will be:
#  * A button to open the view for that schedule item
#  * The button colour will reflect the Conflict status for that schedule item
#     * Due to the inability to change the background colour of a button on a MAC, we will use
#       borders around our button to indicate the various colour statuses of our scheduled items
#
# EVENT HANDLERS:
#    btn_callback(resource)
# ============================================================================
"""

from __future__ import annotations

import platform
from functools import partial
import tkinter as tk
from typing import Callable

from ..Utilities.Colour import is_light
from ..modified_tk import Scrolled, set_default_fonts_and_colours
from ..modified_tk import get_fonts_and_colours
from ..Utilities import Colour
from ..gui_generics.block_colours import RESOURCE_COLOURS
from ..model import ConflictType, ResourceType

class ViewsControllerTk:

    # ============================================================================
    # constructor
    # ============================================================================
    def __init__(self, parent: tk.Frame, resources:dict[ResourceType,list], btn_callback: Callable):
        """
        Initialize this manager
        :param parent: where to put the gui objects
        """

        # remove anything that was already in the frame
        self.parent = parent

        for widget in parent.winfo_children():
            widget.destroy()

        # set fonts
        self.colours, self.fonts = get_fonts_and_colours()
        if self.fonts is None:
            self.colours, self.fonts = set_default_fonts_and_colours(self.parent.winfo_toplevel())
        scrolled_frame = Scrolled(parent, "Frame").widget

        self._button_refs: dict[str, tk.Button] = {}

        tk.Label(scrolled_frame,text="Edit Class Times for ...", font=self.fonts.big, anchor='center').pack(expand=1,fill='both', pady=5)

        # for each resource, create a bunch of buttons that can launch views
        for resource_type in (ResourceType.teacher, ResourceType.lab, ResourceType.stream):
            l_frame = tk.LabelFrame(scrolled_frame, text=resource_type.name)
            l_frame.pack(expand=1,fill='both', padx=5,pady=15)

            for index,resource in enumerate(resources[resource_type]):
                row = int(index/4)
                col = index-row*4
                command = partial(btn_callback, resource)
                self._button_refs[resource.number] = tk.Button(l_frame, text=str(resource), highlightthickness=4, command=command, width=15)
                self._button_refs[resource.number].grid(column = col, row=row, sticky='nsew',ipadx=20, ipady=10, padx=2, pady=2)

    # ============================================================================
    # set button colour
    # ============================================================================
    def set_button_colour(self, button_id: str, resource_type, view_conflict: ConflictType = None):
        """
        set the colour of the button according to the conflict
        :param button_id:
        :param resource_type:
        :param view_conflict:
        """
        btn = self._button_refs.get(button_id, None)
        if btn is None:
            return

        # change the background colour of the button
        # ... (doesn't work on MAC) so also change the background colour of the highlighted area

        colour_highlight = self.parent.cget("background")
        colour = RESOURCE_COLOURS.get(resource_type, self.colours.ButtonBackground)
        active_colour = self.colours.ActiveBackground

        if view_conflict != ConflictType.NONE:
            colour = ConflictType.colours().get(view_conflict,'grey')
            active_colour = Colour.darken(colour, 10)

        if platform.system() == "Darwin":
            colour_highlight = colour
        else:
            if is_light(colour):
                btn.configure(foreground="black")
            else:
                btn.configure(foreground="white")

        btn.configure(background=colour, activebackground=active_colour,
                      highlightbackground=colour_highlight)






from __future__ import annotations
from .enums import ResourceType


class Teacher:
    """Describes a teacher."""

    # -------------------------------------------------------------------
    # constructor
    # --------------------------------------------------------------------
    def __init__(self, firstname: str, lastname: str, department: str = "", release: float = 0):
        """Creates a Teacher object.
        :param firstname:   first name of the teacher
        :param lastname:   last name of the teacher
        :param release: how much release does the teacher have in fractions of FTE
        :param department:  department that this teacher is associated with
        """
        self.firstname = firstname
        self.lastname = lastname
        self.department = department
        self.release = release
        self.resource_type = ResourceType.teacher
        self._id = f"{self.lastname}_{self.firstname}"
        self._id = self._id.replace(" ", "_")

    # -------------------------------------------------------------------------
    # unique identifier
    # -------------------------------------------------------------------------
    @property
    def teacher_id(self) -> str:
        return self._id

    @property
    def number(self) -> str:
        """Returns the unique ID for this Teacher."""
        return self.teacher_id

    # -------------------------------------------------------------------------
    # other
    # -------------------------------------------------------------------------
    def __str__(self) -> str:
        return f"{self.firstname} {self.lastname}"

    def __repr__(self):
        return str(self)

    def __lt__(self, other: Teacher) -> bool:
        return (self.firstname, self.lastname) < (other.firstname, other.lastname)

    def __eq__(self, other):
        return self.number == other.number

    def __hash__(self):
        return hash(self.number)
